;;; FILE: "eg.Music.lisp"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;

(defvar a-music)
(defvar the-music-player)
(defvar the-music-library)

;;;==============================================
;;; class <music> -- holds data for one track
;;;==============================================
(defclass <music> () 
  ( (title        :accessor title        :initarg :title        :type 'string  :initform "")
    (artist       :accessor artist       :initarg :artist       :type 'string  :initform "")
    (album        :accessor album        :initarg :album        :type 'string  :initform "")
    (genre        :accessor genre        :initarg :genre        :type 'string  :initform "")
    (size         :accessor size         :initarg :size         :type 'integer  :initform 0)
    (seconds      :accessor seconds      :initarg :seconds      :type 'integer  :initform 0)
    (track-number :accessor track-number :initarg :track-number :type 'integer  :initform 0)
    (track-count  :accessor track-count  :initarg :track-count  :type 'integer  :initform 0)
    (year         :accessor year         :initarg :year         :type 'integer  :initform 0)
    (date         :accessor date         :initarg :date         :type '<date>)
    (selected     :accessor selected     :initarg :selected     :type 'boolean :initform nil)
  )
)

(setq a-music (make-instance '<music>))

(defmethod print-object ( (m <music>) port )
   (format port "#[<music>TITLE=~s]" (title m)))

(defmethod track ( (m <music>) )
  (format nil "~a of ~a" (track-number m) (track-count m))
)

(defmethod duration ( (m <music>) )
"DURATION is play time in minutes"
  (seconds->minutes (seconds m)))

(defmethod parse-music ( (input string) )
  "return a <music> object"
  (let ( (tab-str (format nil "~a" #\tab)) )
    (labels
	((get-fields (result pos) ;; break up string into list of fields
	    (let ( (tab-pos (search-for tab-str input pos)) )
	      (if (null tab-pos)
		  (reverse result)
		  (get-fields (cons (subseq input pos tab-pos) result)
			      (+ tab-pos 1))))
	 )
	)
      (let*
	  ( (fields (get-fields '() 0))
	    (title  (nth 0 fields))
	    (artist (nth 1 fields))
	    (album  (nth 2 fields))
	    (genre  (nth 3 fields))
	    (size    (string->number (nth 4 fields)))
	    (seconds (string->number (nth 5 fields)))
	    (track-number (string->number (nth 6 fields)))
	    (track-count  (string->number (nth 7 fields)))
	    (year (string->number (nth 8 fields)))
	    (date (string->date   (nth 9 fields)))
	  )
	(make-instance '<music>
		       :title title
		       :artist artist
		       :album album
		       :genre genre
		       :size size
		       :seconds seconds
		       :track-number track-number
		       :track-count track-count
		       :year year
		       :date date)
) ) ) )


;;;==============================================
;;; music-library -- holds <music> objects
;;;==============================================
;; simulate crufty (non-oo) interface
(defstruct music-library looking contents searching )

(setf the-music-library 
   (make-music-library 
       :looking nil :contents '() :searching nil))

(defmethod print-object ( (m music-library) port )
   (format port "#[<music-library>: ~s titles]" (library-total-songs)))

(defun library-load ( name )
  (setf (music-library-contents the-music-library)
	(with-open-file (inp name :direction :input)
	   (do ( (line (progn (read-line inp)(read-line inp nil 'eof)) (read-line inp nil 'eof))
		 (results '())
	       )
	       ((eq line 'eof) (reverse results))
	     (setq results (cons (parse-music line) results)))))
)

(defun library-select ( music )
  (setf (music-library-looking the-music-library) music))

(defun library-total-songs ()
  (length (music-library-contents the-music-library)))

(defun search-library-for (duration-in-seconds)
  (setf (status the-music-player) 'searching)
  (setf (music-library-searching the-music-library) t)
  (setf (next-search-complete the-de-simulator)
	(schedule the-de-simulator duration-in-seconds))
)

;;; completion event
(defun library-search-complete ()
  (when *debug-fit*
	(format t "library-search-complete at ~a seconds~%" 
		(elapsed the-de-simulator)))
  (setf (status the-music-player)
 	(if (playing the-music-player)
 	    'playing
 	    'ready))
  (setf (music-library-searching the-music-library) nil)
)

;;;==============================================
;;; class <browser>
;;;==============================================

(defclass <browser> (<fixture>) ())

(defmethod label-map ( (f <browser>) (label string) )
  (cond
   ((equal label "total songs") "total-songs")
   (T label)
) )

(defmethod music-library-set-selected-track! ( (track-number integer) )
  (setf (music-library-looking the-music-library)
	(nth track-number (music-library-contents the-music-library))))

(defun for-each (proc some-list)
  (iter again ( (a-list some-list) )
	(if (null a-list)
	    nil ;; done
	  (progn
	    (funcall proc (car a-list))
	    (again (cdr a-list))))
) )

(defun library-for-each (proc)
  (for-each proc (music-library-contents the-music-library)))

(defun library-count-selected ()
  (let ( (count 0) )
    (library-for-each 
     #'(lambda (music)
	 (when (selected music) (setf count (add1 count)))))
    count
) )

(defun library-music-selected ()
  (iter loop ( (results '()) (contents (music-library-contents the-music-library)) )
    (cond
     ((null contents)
      (reverse results)
     )
     ((selected (car contents))
      (loop (cons (car contents) results) (cdr contents))
     )
     (T 
      (loop results (cdr contents))))
) )

(defun library-mark-selected-iff (field-key value)
  ;; selected = (field = value)
  ;; I.e. add or remove mark based on value match
  (library-for-each
   #'(lambda (music)
       (setf (selected music) 
	     (equal value (slot-value music field-key)))))
)

(defun library-mark-selected-if (field-key value)
  ;; don't remove mark if set and no value match
  ;; I.e. adds marks, but does not remove any
  (library-for-each
   #'(lambda (music)
       (if (equal value (slot-value music field-key))
	   (setf (selected music) t))))
)

;(defun search-author (value) nil)
;(setf (symbol-function  'search-author)
;      (library-mark-selected-iff 'author value)))

; (defun-syntax define-search
;   (syntax-rules ()
;     ((define-search ?name ?key ?duration) ;=>
;      (define (?name value)
;        (search-library-for ?duration)
;        (library-mark-selected-iff '?key value))
; ) ) )

(defmacro define-search (%name %key %duration)
  `(defun ,%name (value)
     (search-library-for ,%duration)
     (library-mark-selected-iff ',%key value))
)

(define-search search-artist artist 2.3)
(define-search search-album  album  1.1)
(define-search search-genre  genre  0.2)
(define-search search-year   year   0.8)

(defun search-all ()
  (search-library-for 3.2)
  (library-for-each 
   #'(lambda (m) (setf (selected m) t)))
)

(defun mark-all-as (bool)
  (library-for-each
   #'(lambda (music)
       (setf (selected music) bool))))

(defun mark-all-as-selected ()
  (mark-all-as t))

(defun mark-all-as-UNselected ()
  (mark-all-as nil))


;;;==============================================
;;; class browser
;;;==============================================

(defclass <browser> (<fixture>) ())


(defmethod enter-library ( (f <browser>) (cell <parse>) )
  (when *debug-fit*
	(format t "enter-library: about to load ~s~%" (text cell)))
  (library-load (text cell)))

(defmethod check-total-songs ( (f <browser>) )
  (library-total-songs))

(defmethod check-selected-songs ( (f <browser>) )
  (library-count-selected))

(defmethod enter-select ( (f <browser>) (cell <parse>) ) 
  "set 'looking' by nymber"
  (let* ( (track-number (read-number-from-cell cell)) )
    (if (integerp track-number)
	(music-library-set-selected-track! (sub1 track-number)) ;; 1->0 based
	(error (format nil 
		       "enter-select: could not parse track number: ~a"
		       (text cell)))
) ) )

(defmethod check-title ( (f <browser>) ) 
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (trim-whitespace (title selection)))))

(defmethod check-artist ( (f <browser>) ) 
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (artist selection))))

(defmethod check-album ( (f <browser>) ) 
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (album selection))))

(defmethod check-genre ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (genre selection))))

(defmethod check-size ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (size selection))))

(defmethod check-seconds ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (seconds selection))))

(defmethod check-time ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (duration selection))))

(defmethod check-duration ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (duration selection))))

(defmethod check-track-number ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (track-number selection))))

(defmethod check-track-count ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (track-count selection))))

(defmethod check-track ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (track selection))))

(defmethod check-year ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (year selection))))

(defmethod check-date ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (date->string (date selection)))))


(defmethod press-same-album ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (search-album (album selection)))
) )

(defmethod press-same-artist ( (f <browser>) )
  (let ( (selection (music-library-looking the-music-library)) )
    (and selection (search-artist (artist selection)))
) )

(defmethod press-show-all ( (f <browser>) )
  (search-all)
)

;;;==============================================
;;; class fit.ActionFixture eg.music.Browser
;;;==============================================
(defclass eg.music.Browser  (<browser>) ())
(defclass fit.ActionFixture (<action-fixture>) ())

(defmethod label-map ( (f eg.music.Browser) (label string) )
  (cond
   ((string-equal label "total songs")     "total-songs")
   ((string-equal label "selected songs")  "selected-songs")
   ((string-equal label "track number")    "track-number")
   ((string-equal label "track count")     "track-count")
   ((string-equal label "time")            "duration")
   ((string-equal label "same album")      "same-album")
   ((string-equal label "same artist")     "same-artist")
   ((string-equal label "show all")        "show-all")
   ((string-equal label "search complete") "search-complete")
   ((string-equal label "play complete")   "play-complete")
   ((string-equal label "load complete")   "load-complete")
   ((string-equal label "load jam")        "load-jam")
   (T label) ;; NB: did not invoke (next-method)
) )


(defmethod type-map ( (f fit.ActionFixture) (label string) )
  (cond
   ((string-equal label "check-title")        'string)
   ((string-equal label "check-artist")       'string)
   ((string-equal label "check-album")        'string)
   ((string-equal label "check-genre")        'string)
   ((string-equal label "check-size")         'number)
   ((string-equal label "check-seconds")      'number)
   ((string-equal label "check-time")         'number)
   ((string-equal label "check-total-songs")  'integer)
   ((string-equal label "check-track-number") 'integer)
   ((string-equal label "check-track-count")  'integer)
   ((string-equal label "check-track")        'string)
   ((string-equal label "check-date" )        'string)
   (T (call-next-method f label))
) )


;;;==============================================
;;; class fit.RowFixture
;;;==============================================
(defclass fit.RowFixture  (<row-fixture>) ())



;;;==============================================
;;; class <music-player>
;;;==============================================
(defclass <music-player> ()
  ( (playing   :accessor playing
	       :allocation :class :type '<music>
	       :initform nil)
    (paused    :accessor paused ;; seconds
	       :allocation :class :type 'number
	       :initform 0)
    (message   :accessor message
	       :allocation :class :type 'string
	       :initform "")
    (status    :accessor status
	       :allocation :class :type 'symbol
	       :initform 'ready) ;; ready loading playing paused load-jam
) )

(setq the-music-player (make-instance '<music-player>))

(defmethod make ( (mp <music-player>) )
  the-music-player) ;; singelton instance

;;; controls

(defmethod play ( (mp <music-player>) (selection <music>) )
  (when *debug-fit*
	(format t "play ~a~%" mp))
  (if (= 0 (paused mp))
      (let ( (hysteresis (if (playing mp) 0.3 2.5)) )
	  (setf (status mp) 'loading)
	  (setf (playing mp) selection)
	  (setf (next-play-started the-de-simulator)
		(schedule the-de-simulator hysteresis)))
      (progn
	(setf (status mp) 'playing)
	(setf (next-play-complete the-de-simulator)
	      (schedule the-de-simulator (paused mp)))
	(setf (paused mp) 0)
      )
) )

(defmethod pause-player ( (mp <music-player>) )
  (when *debug-fit*
	(format t "pause ~a~%" mp))
  (when (and (playing mp) (= 0 (paused mp)))
	(setf (status mp) 'pause)
	(setf (paused mp)
	      (- (next-play-complete the-de-simulator)
		 (elapsed the-de-simulator)))
	(when *debug-fit*
	      (format t "pause: remaining=~a~%" (paused mp)))
	(setf (next-play-complete the-de-simulator) 0)
  )
)

(defmethod stop ( (mp <music-player>) )
  (when *debug-fit*
	(format t "stop ~a~%" mp))
  (setf (next-play-started    the-de-simulator) 0)
  (setf (next-play-complete   the-de-simulator) 0)
  (setf (elapsed              the-de-simulator) 0)
;;(setf (next-search-complete the-de-simulator) 0)
  (play-completed mp)
)

;;; status

(defmethod seconds-remaining ( (mp <music-player>) )
  (cond
   ((not (= 0 (paused mp))) (paused mp)
   )
   ((playing mp)
    (- (next-play-complete the-de-simulator)
       (elapsed the-de-simulator))
   )
   (T 0)
) )

(defmethod minutes-remaining ( (mp <music-player>) )
  (seconds->minutes (seconds-remaining mp))
)

;;; events

(defmethod play-started ( (mp <music-player>) )
  (when *debug-fit*
	(format t "play-started ~a~%" mp))
  (setf (status mp) 'playing)
  (unless (playing mp)
	  (setf (playing mp)     ;; selected track
		(music-library-looking the-music-library)))
  (setf (next-play-complete the-de-simulator)
	(schedule the-de-simulator (seconds (playing mp))))  ;; track duration
  (when *debug-fit*
	(format t "play-started elapsed=~a complete=~a~%"
		(elapsed the-de-simulator)
		(next-play-complete the-de-simulator)))
)

(defmethod play-completed ( (mp <music-player>) )
  (when *debug-fit*
	(format t "play-completed elapsed=~a ~a~%" (elapsed the-de-simulator) mp))
  (setf (status  mp) 'ready)
  (setf (playing mp) nil)  
)

  

;;;==============================================
;;; class eg.music.Realtime
;;;=============================================
(defclass eg.music.Realtime (<timed-action-fixture>)
  ( (simulator :accessor simulator 
	       :allocation :class
	       :initform (make-instance '<discrete-event-simulator>))
    (player    :accessor player
	       :allocation :class
	       :initform (make-instance '<music-player>))
) )

(defmethod type-map ( (f eg.music.Realtime) (label string) )
  (cond
   ((string-equal label "check-status")         'string)
   ((string-equal label "check-playing")        'string)
   ((string-equal label "check-message")        'string)
   ((string-equal label "check-remaining")      'number)
   ((string-equal label "check-message")        'string)
   ((string-equal label "check-artist")         'string)
   ((string-equal label "check-selected-songs") 'number)
   ((string-equal label "split")                'number)
   (T (call-next-method f label))
) )

(defmethod enter ( (f eg.music.Realtime) )
  (when *debug-fit*
	(format t "enter ~a~%" f))
  (delay-for (simulator f) 0.8)
  (call-next-method f)
)

(defmethod press-play ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "press-play ~a~%" f))
  (delay-for the-de-simulator 1.2)
  (if (music-library-looking the-music-library)
      (play the-music-player (music-library-looking the-music-library))
      (error "press-play: no music selected!")
) )

(defmethod press-pause ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "press-pause ~a~%" f))
  (delay-for the-de-simulator 1.2)
  (pause-player the-music-player)
)

(defmethod pause ( (f eg.music.Browser) 
		   (seconds number) )
  (when *debug-fit*
	(format t "pause ~a ~a~%" seconds f))
  (delay-for the-de-simulator seconds)
)

(defmethod check-status ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "check-status ~a~%" f))
;  (delay-for the-de-simulator 0.2)
  (status the-music-player)
)

(defmethod check-message ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "check-message ~a ~a~%" (message the-music-player) f))
;  (delay-for the-de-simulator 0.2)
  (message the-music-player)
)

(defmethod check-playing ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "check-playing ~a~%" f))
;  (delay-for the-de-simulator 0.2)
  (if (playing the-music-player)
      (title (playing the-music-player))
      "No Selection")
)

(defmethod check-remaining ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "check-remaining ~a~%" f))
;  (delay-for the-de-simulator 0.2)
  (minutes-remaining the-music-player)
)

(defmethod await-search-complete ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "await-search-complete ~a~%" f))
  (wait-search-complete the-de-simulator)
)

(defmethod await-play-complete ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "await-play-complete ~a~%" f))
  (wait-play-complete the-de-simulator)
)

(defmethod press-ok ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "press-ok ~a~%" f))
  (stop the-music-player)
  (setf (status  the-music-player) 'ready)
  (setf (message the-music-player) "")
)

(defmethod fail-load-jam ( (f eg.music.Browser) )
  (when *debug-fit*
	(format t "fail-load-jam ~a~%" f))
  (setf (status  the-music-player) 'load-jam)
  (setf (message the-music-player) "load jamed")
)


;;;==============================================
;;; class eg.music.Display
;;;==============================================
(defclass eg.music.Display (fit.RowFixture) ())

(defmethod get-target-class ( (f eg.music.Display) )
  (find-class '<music>))

(defmethod query ( (f eg.music.Display) )
  (library-music-selected))

 (defmethod type-map ( (f eg.music.Display) (label string) )
   (cond
    ((string-equal label "title")        'string)
    ((string-equal label "artist")       'string)
    ((string-equal label "album")        'string)
    ((string-equal label "genre")        'string)
    ((string-equal label "size")         'number)
    ((string-equal label "seconds")      'number)
    ((string-equal label "time")         'number)
    ((string-equal label "total-songs")  'integer)
    ((string-equal label "track-number") 'integer)
    ((string-equal label "track-count")  'integer)
    ((string-equal label "track")        'string)
    ((string-equal label "date" )        'string)
    (T (call-next-method f label))
) )

(defmethod label-map ( (f eg.music.Display) (label string) )
  (cond
   ((string-equal label "total songs")     "total-songs")
   ((string-equal label "selected songs")  "selected-songs")
   ((string-equal label "track number")    "track-number")
   ((string-equal label "track count")     "track-count")
   ((string-equal label "time")            "duration")
   ((string-equal label "same album")      "same-album")
   ((string-equal label "same artist")     "same-artist")
   (T label) ;; NB: did not invoke (next-method)
) )


;;;==============================================
;;; class fit.Summary
;;;=============================================
(defclass fit.Summary (<summary>) ())


;;; --- E O F --- ;;;
