;;; FILE: "eg.Music.scm"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;


;;;==============================================
;;; class <music> -- holds data for one track
;;;==============================================

(define-class <music> () 
  ( (title        :accessor title        
		  :init-keyword :title        
		  :type 'string  :init-form "")
    (artist       :accessor artist       
		  :init-keyword :artist       
		  :type 'string  :init-form "")
    (album        :accessor album        
		  :init-keyword :album        
		  :type 'string  :init-form "")
    (genre        :accessor genre        
		  :init-keyword :genre        
		  :type 'string  :init-form "")
    (size         :accessor size         
		  :init-keyword :size         
		  :type 'integer  :init-form 0)
    (seconds      :accessor seconds      
		  :init-keyword :seconds      
		  :type 'integer  :init-form 0)
    (track-number :accessor track-number 
		  :init-keyword :track-number 
		  :type 'integer  :init-form 0)
    (track-count  :accessor track-count  
		  :init-keyword :track-count  
		  :type 'integer  :init-form 0)
    (year         :accessor year         
		  :init-keyword :year         
		  :type 'integer  :init-form 0)
    (date         :accessor date         
		  :init-keyword :date         
		  :type <date>)
    (selected     :accessor selected     
		  :init-keyword :selected     
		  :type 'boolean :init-form #f)
    (status       :accessor status
		  :allocation :class
		  :type <symbol> :init-form 'ready)
  )
)

(define music (make <music>))

(define-method write-object ( (m <music>) port )
   (format port "#[<music>TITLE=~s]" (title m)))

(define-method track ( (m <music>) )
  (format #f "~a of ~a" (track-number m) (track-count m))
)

(define-method duration ( (m <music>) ) ;; time
  "seconds->minutes"
  (/ (round (/ (seconds m) 0.6)) 100.0)
)

(define-method parse-music ( (input <string>) )
  "return a music object"
  (when *debug-fit*
	(format #t "(parse-music ~s)~%" input))
  (let ( (tab-str (format #f "~a" #\tab)) )
    (letrec
	((get-fields
	  (lambda (result pos) ;; break up string into list of fields
	    (let ( (tab-pos (search-for tab-str input pos)) )
	      (if (not tab-pos)
		  (reverse result)
		  (get-fields (cons (substring input pos tab-pos) result)
			      (+ tab-pos 1)))))
	 )
	)
      (let*
	  ( (fields  (get-fields '() 0))
	    (title   (nth 0 fields))
	    (artist  (nth 1 fields))
	    (album   (nth 2 fields))
	    (genre   (nth 3 fields))
	    (size    (string->number (nth 4 fields)))
	    (seconds (string->number (nth 5 fields)))
	    (track-number (string->number (nth 6 fields)))
	    (track-count  (string->number (nth 7 fields)))
	    (year         (string->number (nth 8 fields)))
	    (date         (string->date   (nth 9 fields)))
	  )
	(make <music>
	  :title   title
	  :artist  artist
	  :album   album
	  :genre   genre
	  :size    size
	  :seconds seconds
	  :track-number track-number
	  :track-count  track-count
	  :year    year
	  :date    date)
) ) ) )


;;;==============================================
;;; music library -- holds music info objects
;;;==============================================
;; This is a mix of semi-ramdom implementations
;; to simulate interfacing into a hacked design.
;; [Look ma, not objects! 8^]
(define music-library (vector "Music Library" #f '() #f))
(define (music-library-looking)    (vector-ref music-library 1))
(define (music-library-contents)   (vector-ref music-library 2))
(define (music-library-searching?) (vector-ref music-library 3))
(define (music-library-looking-set!  a-music)   
  (vector-set! music-library 1 a-music))
(define (music-library-contents-set! a-music-list)  
  (vector-set! music-library 2 a-music-list))
(define (music-library-search-completed)
  (vector-set! music-library 3 #f))
(define (music-library-start-search)
  (vector-set! music-library 3 #t))


(define (library-load name)
  (music-library-contents-set!
      (call-with-input-file name
	(lambda (inp)
	  (read-line inp) ;; skip 1st (title) line
	  (let loop ( (line (read-line inp)) (results '()) )
	    (if (eof-object? line) 
		(reverse results)
		(loop (read-line inp) (cons (parse-music line) results)))
) ) ) ) )

(define-method library-select ( (music <music>) )
  (music-library-looking-set! music))

(define (search-library-for duration-in-seconds)
  (set! (status music-player) 'searching)
  (music-library-start-search)
  (set! (next-search-complete de-simulator)
	(schedule de-simulator duration-in-seconds))
)

;;; completion event
(define (library-search-complete)

  (when *debug-fit*
	(format #t "library-search-complete at ~a seconds~%" 
		(elapsed de-simulator)))
  (set! (status music-player)
 	(if (playing music-player)
 	    'playing
 	    'ready))
  (music-library-search-completed)
)
  
  

(define (library-total-songs)
  (length (music-library-contents)))

(define-method music-library-set-selected-track! ( (track-number <integer>) )
  (music-library-looking-set!
   (nth track-number (music-library-contents))))

(define (library-for-each proc)
  (for-each proc (music-library-contents)))

(define (library-count-selected)
  (let ( (count 0) )
    (library-for-each 
     (lambda (music)
       (when (selected music) (set! count (add1 count)))))
    count
) )

(define (library-music-selected)
  (let loop ( (results '()) (contents (music-library-contents)) )
    (cond
     ((null? contents)
      (reverse results)
     )
     ((selected (car contents))
      (loop (cons (car contents) results) (cdr contents))
     )
     (else 
      (loop results (cdr contents))))
) )

(define (library-mark-selected-iff field-key value)
  ;; selected = (field = value)
  ;; I.e. add or remove mark based on value match
  (library-for-each
   (lambda (music)
     (set! (selected music) 
	   (equal? value (slot-value music field-key)))))
)

(define (library-mark-selected-if field-key value)
  ;; don't remove mark if set and no value match
  ;; I.e. adds marks, but does not remove any
  (library-for-each
   (lambda (music)
     (if (equal? value (slot-value music field-key))
	 (set! (selected music) #t))))
)


;(define (search-author value)
;  (library-mark-selected-iff 'author value))

(define-syntax define-search
  (syntax-rules ()
    ((define-search ?name ?key ?duration) ;=>
     (define (?name value)
       (search-library-for ?duration)
       (library-mark-selected-iff '?key value))
) ) )

(define-search search-artist artist 2.3)
(define-search search-album  album  1.1)
(define-search search-genre  genre  0.2)
(define-search search-year   year   0.8)

(define (search-all)
  (search-library-for 3.2)
  (library-for-each (lambda (m) (set! (selected m) #t))))


(define (mark-all-as bool)
  (library-for-each
   (lambda (music)
     (set! (selected music) bool))))

(define (mark-all-as-selected)
  (mark-all-as #t))

(define (mark-all-as-UNselected)
  (mark-all-as #f))


;;;==============================================
;;; class browser
;;;==============================================

(define-class <browser> (<fixture>) ())


(define-method enter-library ( (f <browser>) (cell <parse>) )
  (when *debug-fit*
	(format #t "enter-library: about to load ~s~%" (text cell)))
  (library-load (text cell)))

(define-method check-total-songs ( (f <browser>) )
  (library-total-songs))

(define-method check-selected-songs ( (f <browser>) )
  (library-count-selected))

(define-method enter-select ( (f <browser>) (cell <parse>) ) 
  "set 'looking' by nymber"
  (let* ( (track-number (read-number-from-cell cell)) )
    (if (integer? track-number)
	(music-library-set-selected-track! (sub1 track-number)) ;; 1->0 based
	(error (format #f 
		       "enter-select: could not parse track number: ~a"
		       (text cell)))
) ) )

(define-method check-title ( (f <browser>) ) 
  (let ( (selection (music-library-looking)) )
    (and selection (trim-whitespace (title selection)))))

(define-method check-artist ( (f <browser>) ) 
  (let ( (selection (music-library-looking)) )
    (and selection (artist selection))))

(define-method check-album ( (f <browser>) ) 
  (let ( (selection (music-library-looking)) )
    (and selection (album selection))))

(define-method check-genre ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (genre selection))))

(define-method check-size ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (size selection))))

(define-method check-seconds ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (seconds selection))))

(define-method check-time ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (duration selection))))

(define-method check-duration ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (duration selection))))

(define-method check-track-number ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (track-number selection))))

(define-method check-track-count ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (track-count selection))))

(define-method check-track ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (track selection))))

(define-method check-year ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (year selection))))

(define-method check-date ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (date->string (date selection)))))


(define-method press-same-album ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (search-album (album selection)))
) )

(define-method press-same-artist ( (f <browser>) )
  (let ( (selection (music-library-looking)) )
    (and selection (search-artist (artist selection)))
) )

(define-method press-show-all ( (f <browser>) )
  (search-all)
)

;;;==============================================
;;; class fit.ActionFixture eg.music.Browser
;;;==============================================
(define-class eg.music.Browser  (<browser>) ())
(define-class fit.ActionFixture (<action-fixture>) ())

(define-method label-map ( (f eg.music.Browser) (label <string>) )
  (cond
   ((string-ci=? label "total songs")     "total-songs")
   ((string-ci=? label "selected songs")  "selected-songs")
   ((string-ci=? label "track number")    "track-number")
   ((string-ci=? label "track count")     "track-count")
   ((string-ci=? label "time")            "duration")
   ((string-ci=? label "same album")      "same-album")
   ((string-ci=? label "same artist")     "same-artist")
   ((string-ci=? label "show all")        "show-all")
   ((string-ci=? label "search complete") "search-complete")
   ((string-ci=? label "play complete")   "play-complete")
   ((string-ci=? label "load complete")   "load-complete")
   ((string-ci=? label "load jam")        "load-jam")
   (else label) ;; NB: did not invoke (next-method)
) )


(define-method type-map ( (f fit.ActionFixture) (label <string>) )
  (cond
   ((string-ci=? label "check-title")        'string)
   ((string-ci=? label "check-artist")       'string)
   ((string-ci=? label "check-album")        'string)
   ((string-ci=? label "check-genre")        'string)
   ((string-ci=? label "check-size")         'number)
   ((string-ci=? label "check-seconds")      'number)
   ((string-ci=? label "check-time")         'number)
   ((string-ci=? label "check-total-songs")  'integer)
   ((string-ci=? label "check-track-number") 'integer)
   ((string-ci=? label "check-track-count")  'integer)
   ((string-ci=? label "check-track")        'string)
   ((string-ci=? label "check-date" )        'string)
   (else (next-method))
) )


;;;==============================================
;;; class fit.RowFixture
;;;==============================================
(define-class fit.RowFixture  (<row-fixture>) ())




;;;==============================================
;;; class <music-player>
;;;=============================================
(define-class <music-player> ()
  ( (playing   :accessor playing
	       :allocation :class :type <music>
	       :init-form #f)
    (paused    :accessor paused ;; seconds
	       :allocation :class :type <number>
	       :init-form 0)
    (message   :accessor message
	       :allocation :class :type <string>
	       :init-form "")
    (status    :accessor status
	       :allocation :class :type <symbol>
	       :init-form 'ready) ;; ready loading playing paused load-jam
) )

(define music-player (make <music-player>))

(define-method make ( (mp <music-player>) )
  music-player) ;; singelton instance

;;; controls

(define-method play ( (mp <music-player>) (selection <music>) )
  (when *debug-fit*
	(format #t "play ~a~%" mp))
  (if (zero? (paused mp))
      (let ( (hysteresis (if (playing mp) 0.3 2.5)) )
	  (set! (status mp) 'loading)
	  (set! (playing mp) selection)
	  (set! (next-play-started de-simulator)
		(schedule de-simulator hysteresis)))
      (begin
	(set! (status mp) 'playing)
	(set! (next-play-complete de-simulator)
	      (schedule de-simulator (paused mp)))
	(set! (paused mp) 0)
      )
) )

(define-method pause ( (mp <music-player>) )
  (when *debug-fit*
	(format #t "pause ~a~%" mp))
  (when (and (playing mp) (zero? (paused mp)))
	(set! (status mp) 'pause)
	(set! (paused mp)
	      (- (next-play-complete de-simulator)
		 (elapsed de-simulator)))
	(when *debug-fit*
	      (format #t "pause: remaining=~a~%" (paused mp)))
	(set! (next-play-complete de-simulator) 0)
  )
)

(define-method stop ( (mp <music-player>) )
  (when *debug-fit*
	(format #t "stop ~a~%" mp))
  (set! (next-play-started    de-simulator) 0)
  (set! (next-play-complete   de-simulator) 0)
  (set! (elapsed                 de-simulator) 0)
;;(set! (next-search-complete de-simulator) 0)
  (play-completed mp)
)

;;; status

(define-method seconds-remaining ( (mp <music-player>) )
  (cond
   ((not (zero? (paused mp))) (paused mp)
   )
   ((playing mp)
    (- (next-play-complete de-simulator)
       (elapsed de-simulator))
   )
   (else 0)
) )

(define-method minutes-remaining ( (mp <music-player>) )
  (seconds->minutes (seconds-remaining mp))
)

;;; events

(define-method play-started ( (mp <music-player>) )
  (when *debug-fit*
	(format #t "play-started ~a~%" mp))
  (set! (status mp) 'playing)
  (unless (playing mp)
	  (set! (playing mp)     ;; selected track
		(music-library-looking)))
  (set! (next-play-complete de-simulator)
	(schedule de-simulator (seconds (playing mp))))  ;; track duration
  (when *debug-fit*
	(format #t "play-started elapsed=~a complete=~a~%"
		(elapsed de-simulator)
		(next-play-complete de-simulator)))
)

(define-method play-completed ( (mp <music-player>) )
  (when *debug-fit*
	(format #t "play-completed elapsed=~a ~a~%" (elapsed de-simulator) mp))
  (set! (status  mp) 'ready)
  (set! (playing mp) #f)  
)

  

;;;==============================================
;;; class eg.music.Realtime
;;;=============================================
(define-class eg.music.Realtime (<timed-action-fixture>)
  ( (simulator :accessor simulator 
	       :allocation :class
	       :init-form (make <discrete-event-simulator>))
    (player    :accessor player
	       :allocation :class
	       :init-form (make <music-player>))
) )

(define-method type-map ( (f eg.music.Realtime) (label <string>) )
  (cond
   ((string-ci=? label "check-status")         'string)
   ((string-ci=? label "check-playing")        'string)
   ((string-ci=? label "check-message")        'string)
   ((string-ci=? label "check-remaining")      'number)
   ((string-ci=? label "check-message")        'string)
   ((string-ci=? label "check-artist")         'string)
   ((string-ci=? label "check-selected-songs") 'number)
   ((string-ci=? label "split")                'number)
   (else (next-method))
) )

(define-method enter ( (f eg.music.Realtime) )
  (when *debug-fit*
	(format #t "enter ~a~%" f))
  (delay-for (simulator f) 0.8)
  (next-method)
)

(define-method press-play ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "press-play ~a~%" f))
  (delay-for de-simulator 1.2)
  (if (music-library-looking)
      (play music-player (music-library-looking))
      (error "press-play: no music selected!")
) )

(define-method press-pause ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "press-pause ~a~%" f))
  (delay-for de-simulator 1.2)
  (pause music-player)
)


(define-method pause ( (f eg.music.Browser) 
		       (seconds <number>) )
  (when *debug-fit*
	(format #t "pause ~a ~a~%" seconds f))
  (delay-for de-simulator seconds)
)

(define-method check-status ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "check-status ~a~%" f))
;  (delay-for de-simulator 0.2)
  (status music-player)
)

(define-method check-message ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "check-message ~a ~a~%" (message music-player) f))
;  (delay-for de-simulator 0.2)
  (message music-player)
)

(define-method check-playing ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "check-playing ~a~%" f))
;  (delay-for de-simulator 0.2)
  (if (playing music-player)
      (title (playing music-player))
      "No Selection")
)

(define-method check-remaining ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "check-remaining ~a~%" f))
;  (delay-for de-simulator 0.2)
  (minutes-remaining music-player)
)

(define-method await-search-complete ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "await-search-complete ~a~%" f))
  (wait-search-complete de-simulator)
)

(define-method await-play-complete ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "await-play-complete ~a~%" f))
  (wait-play-complete de-simulator)
)

(define-method press-ok ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "press-ok ~a~%" f))
  (stop music-player)
  (set! (status  music-player) 'ready)
  (set! (message music-player) "")
)

(define-method fail-load-jam ( (f eg.music.Browser) )
  (when *debug-fit*
	(format #t "fail-load-jam ~a~%" f))
  (set! (status  music-player) 'load-jam)
  (set! (message music-player) "load jamed")
)


;;;==============================================
;;; class eg.music.Display
;;;==============================================
(define-class eg.music.Display (fit.RowFixture) ())

(define-method get-target-class ( (f eg.music.Display) )
  <music>)

(define-method query ( (f eg.music.Display) )
  (library-music-selected))

 (define-method type-map ( (f eg.music.Display) (label <string>) )
   (cond
    ((string-ci=? label "title")        'string)
    ((string-ci=? label "artist")       'string)
    ((string-ci=? label "album")        'string)
    ((string-ci=? label "genre")        'string)
    ((string-ci=? label "size")         'number)
    ((string-ci=? label "seconds")      'number)
    ((string-ci=? label "time")         'number)
    ((string-ci=? label "total-songs")  'integer)
    ((string-ci=? label "track-number") 'integer)
    ((string-ci=? label "track-count")  'integer)
    ((string-ci=? label "track")        'string)
    ((string-ci=? label "date" )        'string)
    (else (next-method))
) )

(define-method label-map ( (f eg.music.Display) (label <string>) )
  (cond
   ((string-ci=? label "total songs")     "total-songs")
   ((string-ci=? label "selected songs")  "selected-songs")
   ((string-ci=? label "track number")    "track-number")
   ((string-ci=? label "track count")     "track-count")
   ((string-ci=? label "time")            "duration")
   ((string-ci=? label "same album")      "same-album")
   ((string-ci=? label "same artist")     "same-artist")
   (else label) ;; NB: did not invoke (next-method)
) )


;;;==============================================
;;; class fit.Summary
;;;=============================================
(define-class fit.Summary (<summary>) ())


;;; --- E O F --- ;;;
