;;; FILE: "eg.lisp"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;


;;;==============================================
;;; class  eg.ArithmeticFixture
;;;==============================================
(defclass eg.ArithmeticFixture (<fixture>) 
  ( (x :accessor x )
    (y :accessor y ) )
)

(defmethod do-rows ( (f eg.ArithmeticFixture) (rows <parse>) )
  (iter loop ( (row (more rows)) ) ;; skip the 1st row
     (when row
       (do-row f row)
       (loop (more row)))
) )


(defmethod do-cell ( (f eg.ArithmeticFixture) 
		      (cell <parse>) 
		      (column-number integer) )
  (when *debug-fit*
    (format t "cell[~a]=~a~%" column-number (text cell)))
  (case column-number
    ((0) (setf (x f) (read-number-from-cell cell)))
    ((1) (setf (y f) (read-number-from-cell cell)))
    ((2) (check f cell (+ (x f) (y f))))
    ((3) (check f cell (- (x f) (y f))))
    ((4) (check f cell (* (x f) (y f))))
    ((5) (check f cell (/ (x f) (y f))))
    (T (ignore f cell)))
)


 
;;;==============================================
;;; class eg.ArithmeticColumnFixture
;;;==============================================

(defclass eg.ArithmeticColumnFixture (<column-fixture>) 
  ( (x :accessor x )
    (y :accessor y ) )
)

(defmethod plus ( (f eg.ArithmeticColumnFixture) )
  (+ (x f) (y f)))

(defmethod minus ( (f eg.ArithmeticColumnFixture) )
  (- (x f) (y f)))

(defmethod times ( (f eg.ArithmeticColumnFixture) )
  (* (x f) (y f)))

(defmethod divide ( (f eg.ArithmeticColumnFixture) )
  (when *debug-fit*
    (format t "~%divide x=~s, y=~s, x/y=~s~%" (x f) (y f) (/ (x f) (y f))))
  (/ (x f) (y f)))

(defmethod floating ( (f eg.ArithmeticColumnFixture) )
  (+ 0.0 (/ (x f) (y f))))

(defmethod type-map ( (f eg.ArithmeticColumnFixture) (label string) )
  (cond
   ((equal label "floating") 'float)
   (T (call-next-method f label))
) )

;;;==============================================
;;; class eg.Calculator
;;;==============================================

(defclass <hp35-calculator> () 
  ( (registers :accessor registers 
	       :initform (vector 0 0 0 0))
    (store :accessor store :initform 0)
  )
)


(defmethod key ( (calculator <hp35-calculator>) (key string) )
  (cond
   ((key-numeric-p key)
    (do-push calculator (string->number key))
   )
   ((equal key "enter")
    (raise-stack calculator)
   )
   ((equal key "+")
    (do-push calculator (+ (do-pop calculator) (do-pop calculator)))
   )
   ((equal key "-")
    (do-push calculator 
       (let ( (x (do-pop calculator))
	      (y (do-pop calculator))
	    )
	 (- y x)))
   )
   ((equal key "*")
    (do-push calculator (* (do-pop calculator) (do-pop calculator)))
   )
   ((equal key "/")
    (do-push calculator 
       (let ( (x (do-pop calculator))
	      (y (do-pop calculator))
	    )
	 (/ y x)))
   )
   ((equal key "x^y")
    (do-push  calculator (expt (do-pop calculator) (do-pop calculator)))
   )
   ((equal key "clx")
    (let ( (regs (registers calculator)) )
      (setf (aref regs 0) 0)
   ))
   ((equal key "clr")
    (setf (registers calculator) (vector 0 0 0 0))
   )
   ((equal key "chs")
    (let ( (regs (registers calculator)) )
     (setf (aref regs 0) (- (aref regs 0)))
   ))
   ((equal key "x<>y")
    (let ( (regs (registers calculator)) )
      (let ( (reg0 (aref regs 0))
	     (reg1 (aref regs 1))
	   )
	(setf (aref regs 0) reg1)
	(setf (aref regs 1) reg0)
   )))
   ((equal key "r!")
    (let ( (regs (registers calculator)) )
      (setf (aref regs 3) (do-pop calculator))
   ))
   ((equal key "sto")
    (setf (store calculator) (aref (registers calculator) 0))
   )
   ((equal key "rcl")
    (do-push calculator (store calculator))
   )
   ((equal key "sqrt")
    (do-push calculator (sqrt (do-pop calculator)))
   )
   ((equal key "ln")
    (do-push calculator (log (do-pop calculator)))
   )
   ((equal key "sin")
    (do-push calculator (sin (degrees->radians (do-pop calculator))))
   )
   ((equal key "cos")
    (do-push calculator (cos (degrees->radians (do-pop calculator))))
   )
   ((equal key "tan")
    (do-push calculator (tan (degrees->radians (do-pop calculator))))
   )
   (T (error "calculator: can't do key: ~s" key))
) )

(defmethod do-push ( (c <hp35-calculator>) (value number) )
  (raise-stack c)
  (setf (aref (registers c) 0) value)
)

(defmethod raise-stack ( (c <hp35-calculator>) )
  (let ( (regs (registers c)) )
    (setf (aref regs 3) (aref regs 2))
    (setf (aref regs 2) (aref regs 1))
    (setf (aref regs 1) (aref regs 0))
) )

(defmethod do-pop ( (c <hp35-calculator>) )
  (let* ( (regs (registers c)) 
	  (result (aref regs 0))
        )
    (setf (aref regs 0) (aref regs 1))
    (setf (aref regs 1) (aref regs 2))
    (setf (aref regs 2) (aref regs 3))
    result
) )

(defmethod key-numeric-p ( (key string) )
  (numberp (ignore-errors (with-input-from-string (inp key) (read inp))))
)


(defun degrees->radians (deg)
  (/ (* deg pi) 180))

;(defun radians->degrees (rad)
;  @@)

;;;==============================================
;;; class  eg.Calculator
;;;==============================================

(defclass eg.Calculator (<column-fixture>)
  ( (volts :accessor volt  :initform 0)
    (a-key :accessor a-key :initform '"")
    (hp    :accessor hp    :initform (make-instance '<hp35-calculator>))
  )
)

(defmethod label-map ( (f eg.Calculator) (label string) )
  (if (equal label "key")
      "a-key"
      label)
)

(defmethod type-map ( (f eg.Calculator) (label string) )
  (cond
   ((equal label "points") 'boolean)
   ((equal label "flash")  'boolean)
   ((equal label "key")    'string)
   (T (call-next-method f label))
) )

(defmethod points ( (fixture eg.Calculator) )
  nil)

(defmethod flash ( (fixture eg.Calculator) )
  nil)

(defmethod watts ( (fixture eg.Calculator) )
  0.5)

(defmethod reset ( (fixture eg.Calculator) )
  (setf (a-key fixture) nil))

(defmethod do-execute ( (fixture eg.Calculator) )
  (and (a-key fixture)
       (key (hp fixture) (a-key fixture))
) )

(defmethod x ( (fixture eg.Calculator) )
  (do-execute fixture)
  (aref (registers (hp fixture)) 0)
)

(defmethod y ( (fixture eg.Calculator) )
  (aref (registers (hp fixture)) 1)
)

(defmethod z ( (fixture eg.Calculator) )
  (aref (registers (hp fixture)) 2)
)

(defmethod t ( (fixture eg.Calculator) )
  (aref (registers (hp fixture)) 3)
)


;;; --- E O F --- ;;;
