;;; FILE: "fit-tests.scm"
;;; IMPLEMENTS: Framework for Integrated Test
;;;  See http://fit.c2.com for documentation and examples
;;;
;;; Copyright (C) 2002 by Ken Dickey & Ward Cunningham
;;;  This is Free Software, covered by the GNU GPL (v2)
;;;  See http://www.gnu.org/copyleft/gpl.html
;;;
;;; DATE:    15 October 2002 -- started coding -- KenD
;;; HISTORY: @@@
;;;
;;; NOTES:
;;;  The code follows the structure of the Java FIT tests

; set! #t to avoid breaking into the debugger
(define treat-errors-as-warnings #t)

;; Light weight unit testing.. only reports on error (@@FIXME@@)

(define-syntax expect
  (syntax-rules ()
    ((expect ?expected ?form) ;=>
     (expect ?expected ?form equal?)
    )
    ((expect ?expected ?form ?compare) ;=>
     (let ( (expected ?expected)
            (actual   (with-error-handler 
		       (lambda (error-obj) (error-string error-obj))
		       (lambda () ?form)))
          )
       (if (?compare expected actual)
	   (format #t "PASSED:  ~s~%" ?form)
	   ((if treat-errors-as-warnings warn error)
	    (format #f 
		    "~%**FAIL: expected: ~s~% got: ~s~% from: ~s~%"
		    expected
		    actual
		    '?form))))
    )
) )

(define-syntax test-assert-equal ;; just an alias
  (syntax-rules ()
    ((test-assert-equal expected actual) ;=>
     (expect expected actual))
) )


;;============================================

(define testParsing (do-parse "leader<Table foo=2>body</table>trailer" '("table")));)

(test-assert-equal  "leader"        (leader  testParsing))
(test-assert-equal  "<Table foo=2>" (tag     testParsing))
(test-assert-equal  "body"          (body    testParsing))
(test-assert-equal  "trailer"       (trailer testParsing))

;;================================================

(define testRecursing 
  (do-parse "leader<table><TR><Td>body</tD></TR></table>trailer" 
	    '("table" "tr" "td")))

(test-assert-equal #f     (body testRecursing))
(test-assert-equal #f     (body (parts testRecursing)))
(test-assert-equal "body" (body (parts (parts testRecursing))))

;;================================================

(define testIterating 
  (do-parse "leader<table><tr><td>one</td><td>two</td><td>three</td></tr></table>trailer"
	    '("table" "tr" "td")))

(test-assert-equal "one"   (body (parts (parts testIterating))))
(test-assert-equal "two"   (body (more (parts (parts testIterating)))))
(test-assert-equal "three" (body (more (more (parts (parts testIterating))))))

;;===============================================

(define testIndexing
  (do-parse 
   "Leader<table><tr><td>one</td><td>two</td><td>three</td></tr><tr><td>four</td></tr></table>trailer"
   '("table" "tr" "td")))

(test-assert-equal "one"   (body (at testIndexing 0 0 0)))
(test-assert-equal "two"   (body (at testIndexing 0 0 1)))
(test-assert-equal "three" (body (at testIndexing 0 0 2)))
(test-assert-equal "three" (body (at testIndexing 0 0 3)))
(test-assert-equal "three" (body (at testIndexing 0 0 4)))
(test-assert-equal "four"  (body (at testIndexing 0 1 0)))
(test-assert-equal "four"  (body (at testIndexing 0 1 1)))
(test-assert-equal "four"  (body (at testIndexing 0 2 0)))
(test-assert-equal 1       (size testIndexing))
(test-assert-equal 2       (size (parts testIndexing)))
(test-assert-equal 3       (size (parts (parts testIndexing))))
(test-assert-equal "one"   (body (leaf testIndexing)))
(test-assert-equal "four"  (body (leaf (last-of (parts testIndexing)))))


;;=======================================================

;; +root-path+ is defined in file "load-me.scm"
(define html-file-name 
  (string-append +root-path+ 
		 "/Release/Documents/" 
		 "arithmetic.html"))

(define fileParsing 
  (do-parse 
   (call-with-input-file html-file-name (lambda (inport) (port->string inport)))
   '("table" "tr" "td")))

(define simple-fixture (make <fixture>))

;; process the tables
(do-tables simple-fixture fileParsing)

(define result-file-name "/tmp/result-from-scheme.html")

(call-with-output-file result-file-name
   (lambda (outport)
     (print-to-stream fileParsing outport)
     (format #t "Check file ~a for results~%" result-file-name)))


;;===============================================

(define tags '("td"))

(define p0 (do-parse "<td>a&lt;b</td>" tags))
(test-assert-equal "a&lt;b" (body p0))
(test-assert-equal "a<b"    (text p0))

(define p1 
  (do-parse "<td>\ta&gt;b&nbsp;&amp;&nbsp;b>c &&&nbsp;</td>" tags))
(test-assert-equal "a>b & b>c &&"  (text p1))

(define p2 
  (do-parse 
   "<TD><P><FONT FACE=\"Arial\" SIZE=2>GroupTestFixture</FONT></TD>"
   tags))

(test-assert-equal "GroupTestFixture" (text p2))

(test-assert-equal "a<b" 
		   (unescape "a&lt;b"))

(test-assert-equal "a>b & b>c &&" 
		   (unescape "a&gt;b&nbsp;&amp;&nbsp;b>c &&"))

(test-assert-equal "ab" 
		   (unformat "<font size=+1>a</font>b"))

(test-assert-equal "ab"  
		   (unformat "a<font size=+1>b</font>"))

(test-assert-equal "a<b" 
		   (unformat "a<b"))


;;===============================================

(define s 
  (do-parse 
   "Leader<table><tr><td>one</td><td>two</td><td>three</td></tr><tr><td>four</td></tr></table>trailer"
   '("table" "tr" "td")))

(test-assert-equal "one"   (body (at s 0 0 0)))
(test-assert-equal "two"   (body (at s 0 0 1)))
(test-assert-equal "three" (body (at s 0 0 2)))
(test-assert-equal "three" (body (at s 0 0 3)))
(test-assert-equal "three" (body (at s 0 0 4)))
(test-assert-equal "four"  (body (at s 0 1 0)))
(test-assert-equal "four"  (body (at s 0 1 1)))
(test-assert-equal "four"  (body (at s 0 2 0)))
(test-assert-equal 1       (size s))
(test-assert-equal 2       (size (parts s)))
(test-assert-equal 3       (size (parts (parts s))))
(test-assert-equal "one"   (body (leaf s)))
(test-assert-equal "four"  (body (leaf (last-of (parts s)))))

(define latest-parse #f)

(define (make-html-page-test outfile-prefix tags table-fun)
  (lambda (base-file-name)
    (format #t "~%==>>processing ~s~%" base-file-name) ;;@@@debug
    (let* ( (in-name  (string-append +root-path+ "/Release/Documents/" base-file-name))
	    (out-name (string-append outfile-prefix  base-file-name))
	    (tables   (do-parse (file->string in-name) tags))
	    (fixture  (make <fixture>))
          )
      (reset-counts fixture)
      (do-tables fixture (table-fun tables))
      (when *debug-fit*
	    (set! latest-parse tables))
      (call-with-output-file out-name
	(lambda (outp) (print-to-stream tables outp)))
      (format #t "~%    ==>> Check file ~a for results~%" out-name)
) ) )

;(set! *debug-fit* #t)

(define test-html-page 
  (make-html-page-test "/tmp/" 
		       '("table" "tr" "td")
		       (lambda (t) t))
)

(define test-wiki-page
  (make-html-page-test "/tmp/" 
		       '("wiki" "table" "tr" "td") 
		       (lambda (t) (parts t)))
)


;;=======================================================
;; Arithmetic test

;; +root-path+ is defined in file "load-me.lisp"

(test-html-page "arithmetic.html")


;;=======================================================
;; Calculator test

(test-wiki-page "CalculatorExample.html")


;;=======================================================
;; Music test

(test-wiki-page "MusicExample.html")

;(set! *debug-fit* #t) ;;@@@debug
(test-wiki-page "MusicExampleWithErrors.html")

'DONE

;;; --- E O F --- ;;;
