;;; FILE: "fit.lisp"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;


;;;  This code parses HTML tables to obtain information on what to test.
;;;  Tables use one of 3 patterns:
;;;	Row     - arg... expected..
;;;	Column  - arg... function  -- may add columns (e.g. db query w extra results)
;;;	Actions - action arg... expected... -- series of action steps
;;;     TimedAction - as with action, but adds timing info
;;;
;;;  This code parses an input html page, 
;;;	processes the tests,  and 
;;;	outputs a html page of results.
;;;
;;;  The code is not speed critical (fast execution is not a requirement).

(eval-when (compile load eval)
;  (require :cllib-html    (translate-logical-pathname "cllib:html"))
;  (require :cllib-htmlgen (translate-logical-pathname "cllib:htmlgen"))
)

(in-package :user)

;(export
; '())


;;;==============================================
;;; Color tag info

(defconstant +wrong-tag-string+      " bgcolor=#ffcfcf")
(defconstant +exception-tag-string+  " bgcolor=#ffffcf")
(defconstant +right-tag-string+      " bgcolor=#cfffcf")
(defconstant +ignore-tag-string+     " bgcolor=#efefef")
(defconstant +gray-tag-string+       " color=#808080")
(defconstant +label-tag-string+      " color=#c08080")

;;;==============================================
;;; Various helper functions

(defparameter *debug-fit* nil)
(defvar the-de-simulator)
(defvar the-music-player)
(defvar a-fixture)

(defun add1 (n) (+ n 1))
(defun sub1 (n) (- n 1))
(defun quotient  (n1 n2) (truncate (/ n1 n2)))
(defun remainder (n1 n2) (rem n1 n2))

;; kinda hard to apply (CONCATENATE 'STRING ...)
(defun string-append ( &REST strings )
  (if (null strings)
      ""
    (do ( (result (car strings) (concatenate 'string result (car others)))
	  (others (cdr strings) (cdr others))
	)
	((null others) result)
) ) )

(defmacro iter (%name %binds . %body)
"(ITER name ( (var init) .. ) . body)
= Named LET"
  (let ( (vars  (mapcar #'car %binds))
	 (inits (mapcar #'cadr %binds))
       )
  `(labels ( (,%name ,vars ,@%body) )
     (,%name ,@inits))
) )

(defun read-file-into-string ( fname )
"(READ-FILE-INTO-STRING string-file-name)
reads the indicated file and returns a string."
  (let ( (input (open fname :direction :input))
	 (endline (format nil "~%"))
       )
    (iter read-loop ( (lines "") (next (ignore-errors (read-line input))) )
	  (if next
	      (read-loop (concatenate 'string lines next endline )
			 (ignore-errors (read-line input)))
	     lines)
    )
) )



(defmacro catch-errors (%a-cell . %forms-to-eval)
  (let ( (catch-tag (gensym)) )
    `(block ,catch-tag
       (handler-bind 
	((condition
	  #'(lambda (condition) 
	      (when *debug-fit*
		(format t 
			"~%***error ~a~%***for cell~% ~s~%" 
			condition 
			,%a-cell)
;		(and ,%a-cell (describe ,%a-cell))
		(let ( (msg (format nil "~a~%" condition)) )
		  (format t msg)
;		  (if (search-for "STRING-APPEND" msg 0)
;		      (invoke-debugger condition))
		))
		(exception ,%a-cell condition)
		(return-from ,catch-tag nil))
	  ))
	 (progn ,@%forms-to-eval)))
) )


(defmethod search-for ( (want string) 
			(to-search string) 
			(start-pos integer) )
"(SEARCH-FOR want to-search start-pos)
searches string TO-SEARCH starting at START-POS 
looking for a match to string WANT
returns the match starting offset or NIL."
  (let ( (want-len      (length want))
	 (to-search-len (length to-search))
       )
    (labels
	((TRY (pos)
	      (let ( (part (subseq to-search pos (+ pos want-len))) )
;		(format t "offset: ~a substr: \"~a\"~%" pos part)
		(string-equal part want))
	 )
	 (ITERATE (pos)
	   (cond
	    ((> (+ pos want-len) to-search-len)
	     nil)
	    ((try pos) pos)
	    (T (iterate (+ 1 pos))))
	)
       )
      (iterate start-pos))))

;(compile 'search-for) ;; make it iterative!!


;;;==============================================
;;; class <date>
;;;==============================================
(defclass <date> ()
  ( (year    :accessor year    :initarg :year    
	     :type 'integer :initform 0)
    (month   :accessor month   :initarg :month   
	     :type 'integer :initform 0)
    (day     :accessor day     :initarg :day     
	     :type 'integer :initform 0)
    (hours   :accessor hours   :initarg :hours   
	     :type 'integer :initform 0)
    (minutes :accessor minutes :initarg :minutes 
	     :type 'integer :initform 0)
    (seconds :accessor seconds :initarg :seconds 
	     :type 'integer :initform 0)
) )

(defmethod clone ( (d <date>) )
  (make-instance '<date>
    :year    (year  d)
    :month   (month d)
    :day     (day   d)
    :hours   (hours d)
    :minutes (minutes d)
    :seconds (seconds d)
) )

(defun make-current-date ()
  (multiple-value-bind (sec min h d m y dw dst-p zone)
      (get-decoded-time)
      (declare (ignore dw dst-p zone))
      (make-instance '<date>
	:year  y
	:month m
	:day   d
	:hours h
	:minutes min
	:seconds sec))
)

(defun seconds->minutes (seconds)
	(+ 0.0 (/ (round (* (/ seconds 60) 100)) 100)))

(defun minutes->seconds (minutes)
  (* minutes 60))

(defun hours->seconds (hours)
  (* 60 60 hours))


(defmethod add ( (d1 <date>) (d2 <date>) )
  ;;@@@ FIXME time only
  (let* ( (s1 (+ (hours->seconds   (hours d1))
		 (minutes->seconds (minutes d1))
		 (seconds d1)))
	  (s2 (+ (hours->seconds   (hours d2))
		 (minutes->seconds (minutes d2))
		 (seconds d2)))
	  (delta (+ s1 s2))
	  (hours   (quotient  delta (* 60 60)))
	  (rem     (remainder delta (* 60 60)))
	  (minutes (quotient  rem 60))
	  (seconds (remainder rem 60))
	)
       (make-instance '<date>
	 :year  (max (year d1)  (year d2))
	 :month (max (month d1) (month d2))
	 :day   (+   (day d1)   (day d2))
	 :hours hours :minutes minutes :seconds seconds)
) )

(defmethod sub ( (d1 <date>) (d2 <date>) )
  ;;@@@ FIXME time only
  (let* ( (s1 (+ (hours->seconds (hours d1))
		 (minutes->seconds (minutes d1))
		 (seconds d1)))
	  (s2 (+ (hours->seconds (hours d2))
		 (minutes->seconds (minutes d2))
		 (seconds d2)))
	  (delta (- s1 s2))
	  (hours   (quotient delta (* 60 60)))
	  (rem     (remainder delta (* 60 60)))
	  (minutes (quotient  rem 60))
	  (seconds (remainder rem 60))
	)
       (make-instance '<date>
	 :hours hours :minutes minutes :seconds seconds)
) )

(defmethod print-object ( (d <date>) port )
  (format port "~a" (date->string d)))

(defmethod string->date ( (str string) )
  ;; sample format: "9/9/02 9:01 AM" D/M/Y h:m AP
  ;; @@ FIXME -- not forgiving at all
 (let* ( (str         (trim-whitespace str))
	 (day-end     (search-for "/" str 0))
	 (month-end   (and day-end 
			   (search-for "/" str (add1 day-end))))
	 (year-end    (and month-end
			   (search-for " " str (add1 month-end))))
	 (hours-end   (and year-end
			   (search-for ":" str (add1 year-end))))
	 (minutes-end (and hours-end
			   (search-for " " str (add1 hours-end))))
	 (am/pm       (and minutes-end
			   (aref str (+ minutes-end 1))))
	)
   (unless minutes-end
	   (error (format nil "string->date: could not parse: ~s" str)))
;    (when *debug-fit*
; 	 (format t "d#~a, m#~a, y#~a, h#~a, m#~a, ~am~%" 
; 		 day-end month-end year-end 
; 		 hours-end minutes-end 
; 		 am/pm))
   (let ( (day     (string->number (subseq str 0 day-end)))
	  (month   (string->number 
		    (subseq str (add1 day-end) month-end)))
	  (year    (string->number 
		    (subseq str (add1 month-end) year-end)))
	  (hours   (string->number 
		    (subseq str (add1 year-end) hours-end)))
	  (minutes (string->number 
		    (subseq str (add1 hours-end) minutes-end)))
	)
;      (when *debug-fit*
; 	 (format t "d=~a, m=~a, y=~a, h=~a, m=~a, ~am~%" 
; 		 day month year 
; 		 hours minutes 
; 		 am/pm))
     (when (member am/pm '(#\p #\P))
	   (setf hours (+ hours 12)))
     (make-instance '<date> :year year :month month :day day 
	          :hours hours :minutes minutes)
) ) )

(defmethod date->string ( (d <date>) )
;; D/M/Y h:m AP
  (let* ( (month   (month d))
	  (day     (day d))
	  (year    (year d))
	  (display-year ;; NB: 2 digits
	   (format nil (if (< year 10) "0~a" "~a") year))
	  (minutes (minutes d))
	  (hours   (hours d))
	  (am/pm 
	   (if (> hours 12) #\P #\A))
	  (display-hours 
	   (if (> hours 12) (- hours 12) hours))
        )
  (format nil "~a/~a/~a ~a:~a~a ~aM" 
	  day month display-year display-hours 
	  (if (< minutes 10) "0" "") minutes am/pm)
) )

(defmethod number->string ( (n number) )
  (format nil "~a" n))

(defmethod date->hms ( (d <date>) )
  ;; hh:mm:ss
  (let ( (h (number->string (hours   d)))
	 (m (number->string (minutes d)))
	 (s (number->string (seconds d)))
	 (digitize 
	   #'(lambda (s)
	       (if (< (length s) 2)
		   (concatenate 'string "0" s)
		   s))
	 )
       )
    (format nil "~a:~a:~a"
	    (funcall digitize h)
	    (funcall digitize m)
	    (funcall digitize s))
) )



;;;==============================================
;;; class <discrete-event-simulator>
;;;=============================================
;;; This discrete event simulator supports three events
;;; each of which is open coded in the body of the simulator.

;;; internal time in absolute seconds
(defclass <discrete-event-simulator> ()
  ( (birth-date :accessor birth-date :type '<date>
		:allocation :class
	        :initarg :birth-date
		)
    (elapsed   :accessor elapsed :type 'number ;; seconds
	       :allocation :class 
	       :initform 0)
    (next-search-complete :accessor next-search-complete 
			  :type 'number ;; seconds
			  :allocation :class :initform 0)
    (next-play-started    :accessor next-play-started 
			  :type 'number ;; seconds
			  :allocation :class :initform 0)
    (next-play-complete   :accessor next-play-complete 
			  :type 'number ;; seconds
			  :allocation :class :initform 0)
) )

;; define singleton simulator
(setq the-de-simulator 
      (make-instance '<discrete-event-simulator>
		     :birth-date (make-current-date)))

(defmethod make ( (ds <discrete-event-simulator>) )
  the-de-simulator) ;; NB: singleton instance

(defmethod sooner ( (ds <discrete-event-simulator>)
		    (soon  number)
		    (event number) )
;; NB: soon must be > elapsed
  (when (< soon (elapsed ds))
	(error "sooner: can't allow negative time"))
  (if (> soon event (elapsed ds)) 
      event
      soon)
)

(defmethod next-event ( (ds <discrete-event-simulator>) 
			(bound number) )
;; NB: assert (<= elapsed bound)
  (let ( (next
	  (sooner ds 
		  (sooner ds 
			  (sooner ds
				  bound
				  (next-search-complete ds))
			  (next-play-started ds))
		  (next-play-complete ds)))
       )
    (when *debug-fit*
	  (format t "next-event: ~a~%" next))
    next
) )

;; send events at specified times
(defmethod perform ( (ds <discrete-event-simulator>) )
  (let ( (time (elapsed ds))
	 (search-done (next-search-complete ds))
	 (play-start  (next-play-started    ds))
	 (play-end    (next-play-complete   ds))
       )
    (when (and (> search-done 0)
	       (>= time search-done))
	  (setf (next-search-complete ds) 0)
	  (library-search-complete))
    (when (and (> play-start 0)
	       (>= time play-start))
	  (setf (next-play-started ds) 0)
	  (play-started the-music-player))
    (when (and (> play-end 0)
	       (>= time play-end))
	  (setf (next-play-complete ds) 0)
	  (play-completed the-music-player))
) )

(defmethod advance ( (ds <discrete-event-simulator>) 
		     (future number) )
  (while (< (elapsed ds) future)
	 (when *debug-fit*
	       (format t 
		       "advance: elapsed=~a future=~a~%" 
		       (elapsed ds) 
		       future)
	       (format t "advance: next-event=~a~%"
		       (next-event ds future)))
	 (setf (elapsed ds)
	       (next-event ds future)
	 )
	 (perform ds))
)

(defmethod schedule ( (ds <discrete-event-simulator>)
		      (seconds number) )
"returns seconds from 'now'"
  (+ (elapsed ds) seconds)
)

(defmethod delay-for ( (ds <discrete-event-simulator>)
		       (seconds number) )
  (advance ds (schedule ds seconds))
)

(defmethod wait-search-complete ( (ds <discrete-event-simulator>) )
  (advance ds (next-search-complete ds))
)

(defmethod wait-play-started ( (ds <discrete-event-simulator>) )
  (advance ds (next-play-started ds))
)

(defmethod wait-play-complete ( (ds <discrete-event-simulator>) )
  (advance ds (next-play-complete ds))
)

(defmethod get-time ( (ds <discrete-event-simulator>) )
;; return a date object
  (let* ( (time    (elapsed ds))
	  (minutes (truncate (seconds->minutes time)))
	  (seconds (truncate (- time (minutes->seconds minutes))))
	  (delta (make-instance '<date> :minutes minutes :seconds seconds))
	)
    (add delta (birth-date ds))
) )


;;;==============================================
;;; class parse
;;;==============================================
		
;;; <PARSE> holds the results of a parse of html text.
(defclass <parse> () 
  ( (leader  :initarg :leader  :accessor leader  :initform "") 
    (start-tag :initarg :start-tag :accessor start-tag) ;; must be specified
    (body    :initarg :body    :accessor body    :initform nil)
    (parts   :initarg :parts   :accessor parts   :initform nil)
    (more    :initarg :more    :accessor more    :initform nil)
    (end-tag :initarg :end-tag     :accessor end-tag) ;; must be specified
    (trailer :initarg :trailer :accessor trailer :initform "")
) )

(defmethod DO-PARSE ( (a-string string) (tag-list list) )
  (let* ( (tag            (car tag-list))
	  (next-tag-list  (cdr tag-list))
	  (start-tag  (concatenate 'string "<"  tag))
	  (end-tag    (concatenate 'string "</" tag))
	  (start-pos  (search-for start-tag a-string 0))
	  (start-end  (and start-pos (search-for ">"       a-string start-pos)))
	  (end-pos    (and start-end (search-for end-tag   a-string start-end)))
	  (end-end    (and end-pos   (search-for ">"       a-string end-pos)))
	  (next-start (and end-end   (search-for start-tag a-string end-end)))
        )
    (when (not end-end)
      (error "do-parse: can't find tag ~a" tag))
    (incf start-end)
    (incf end-end)
    (let ( (body    (subseq a-string start-end end-pos))
	   (trailer (subseq a-string end-end   (length a-string)))
	 )

      (make-instance '<parse> 
		     :leader    (subseq a-string 0         start-pos) 
		     :start-tag (subseq a-string start-pos start-end)
		     :body      (if (null next-tag-list)
				    body
				    nil)
		     :parts     (if (not (null next-tag-list))
				    (do-parse body next-tag-list))
		     :more      (if next-start
				    (do-parse trailer tag-list)
				    nil)
		     :end-tag   (subseq a-string end-pos   end-end) 
		     :trailer   (if (not next-start) trailer nil))
) ) )

(defmethod size ( (p <parse>) )
  (if (more p) (+ 1 (size (more p))) 1))

(defmethod last-of ( (p <parse>) )
  (if (more p) (last-of (more p)) p))

(defmethod leaf ( (p <parse>) )
  (if (parts p) (leaf (parts p)) p))

(defmethod at1 ( (p <parse>) (i integer) )
  (if (or (= i 0) (null (more p))) p (at1 (more p) (- i 1))))

(defmethod at2 ( (p <parse>) (i integer) (j integer) )
  (at1 (parts (at1 p i)) j))

(defmethod at3 ( (p <parse>) (i integer) (j integer) (k integer) )
  (at1 (parts (at2 p i j)) k))

(defmethod add-to-tag ( (p <parse>) (text string) )
  (setf (start-tag p)
	(concatenate 'string 
		     (subseq (start-tag p) 0 (- (length (start-tag p)) 1)) 
		     text 
		     ">")))

(defmethod add-to-body ( (p <parse>) (text string) )
  (setf (body p)
	(concatenate 'string (body p) text)))

;;; PRINT-TO-STREAM prints the <parse> P back into HTML
(defmethod print-to-stream ( (p <parse>) &optional (stream *standard-output*) )
  (when (leader p)
    (format stream "~a" (leader p)))
  (format stream "~a" (start-tag p))
  (when (body p)
    (format stream "~a" (body p)))
  (when (parts p)
    (print-to-stream (parts p) stream))
  (format stream "~a" (end-tag p))
  (when (more p)
    (print-to-stream (more p) stream))
  (when (trailer p)
    (format stream "~a" (trailer p)))
  p
)

(defmethod text ( (p <parse>) )
"(TEXT <parse>)
returns the string text less tags and whitespece"
  (trim-whitespace (unescape (unformat (body p)))))

(defmethod unformat ( (str string) )
  (iter again ( (str str) )
     (let ( (more (search-for "<" str 0)) )
       (if more
	   (let ( (tag-start more)
		  (tag-end (search-for ">" str (+ more 1)))
		)
	     (if tag-end
		 (again (concatenate 'string 
				    (subseq str 0 tag-start)
				    (subseq str (+ 1 tag-end) (length str))))
	         str))
	   str)
) ) )


(defmethod unescape ( (str string) )
  (labels
      (
        (SUNSTITUTE (esc)
	   (when *debug-fit*
	     (format t "sunstitute for ~s~%" esc))
	   (cond
	    ((equal esc "&lt;")   "<")
	    ((equal esc "&gt;")   ">")
	    ((equal esc "&amp;")  "&")
	    ((equal esc "&nbsp;") " ")
	    ((equal esc "&quot;") "\"")
	    (T nil))
        )
	(ITERATE (str pos)
	 (let ( (more (search-for "&" str pos)) )
	  (if more
	      (let* ( (esc-start more)
		      (esc-end   (and more (search-for ";" str (add1 esc-start))))
		      (len       (length str))
		      (double-amp-p (and esc-start
					 (< (add1 esc-start) len)
					 (equal #\& (aref str (add1 esc-start)))))
		    )
;		(format t "escape from=~a to=~a~%" esc-start esc-end) ;;@@@
		(cond
		 (double-amp-p ; "&&" -> skip first "&"
		  (iterate str (add1 esc-start))
		 )
		 (esc-end
		  (let* ( (subst-piece (or (sunstitute 
					    (subseq str 
						    esc-start 
						    (+ esc-end 1)))
					   ""))
			  (next-pos (+ esc-start (length subst-piece)))
			  )
		    (iterate (concatenate 'string 
				       (subseq str 0 esc-start)
				       subst-piece
				       (subseq str (+ 1 esc-end) (length str)))
			  next-pos))
		  )
		  (T str)))
	      str)
	  ))
	)
       (iterate str 0)
) )


(defmethod trim-whitespace ( (str string) )
  (let ( (start 0)
	 (len (length str))
       )
    (labels
	((GET-START (cur)
	   (cond
	     ((>= cur len) "")
	     ((whitespace-p (aref str cur))
	      (setq start (+ cur 1))
	      (get-start (+ cur 1)))
	     (T (get-end (- len 1))))
	  )
	 (GET-END (cur)
	   (cond
	     ((< cur 0) (subseq str start (+ cur 1)))
	     ((whitespace-p (aref str cur))
	      (get-end (- cur 1)))
	     (T (subseq str start (+ cur 1))))
	  )
	 )
      (get-start 0)
) ) )

(defun whitespace-p ( ch )
    (member ch '( #\space #\tab #\newline ) ) )

(defmethod exception ( (p <parse>) (c condition) )
;;  (invoke-debugger c)
  (incf (num-excepts a-fixture))
  (add-to-body p (format nil "<hr>~a" c))
  (add-to-tag p +exception-tag-string+))

;;;==============================================
;;; class <fixture>
;;;==============================================

(defclass <fixture> () 
  ( (num-right   :accessor num-right    :allocation :class :initform 0)
    (num-wrong   :accessor num-wrong    :allocation :class :initform 0)
    (num-ignores :accessor num-ignores  :allocation :class :initform 0)
    (num-excepts :accessor num-excepts  :allocation :class :initform 0)
    (summary    :accessor summary       :allocation :class :initform '())
) )

(setq a-fixture (make-instance '<fixture>))

(defmethod do-tables ( (f <fixture>) (tables <parse>) )
  (iter again ( (table tables) )
     (when table
       (let ( (cell (at3 table 0 0 0)) )
	 (catch-errors 
	  cell
	  (let ( (fixture-name (and table (read-fixture cell))) )
	    (do-table (make-instance fixture-name) table)))
	 (again (more table))
     ) )
) )

(defun string->symbol ( a-string )
  (with-input-from-string (inp a-string) (read inp)))

(defmethod string->number ( (str string) )
  (let ( (value (ignore-errors 
		  (with-input-from-string (inp str) (read inp))))
       )
    (if (numberp value)
	value
        (error "string->number: not a number " str))
) )

(defmethod read-fixture ( (cell <parse>) )
"(READ-FIXTURE <parse>)
reads a class-name from the 1st cell in a <parse> instance
or throws an error;
returns a symbol"
  (string->symbol (text cell))
)

(defmethod do-table ( (f <fixture>) (table <parse>) )
;;  (format t "DO-TABLE called~%") ;; @@@
  (do-rows f (more (parts table)))) ;; skip fixture name

(defmethod do-rows ( (f <fixture>) (rows <parse>) )
  (iter again ( (row rows) )
     (when row
       ;; be careful of rows added by processing
       (let ( (next (more row)) )
	 (do-row f row)
	 (again next))
) ) )

(defmethod do-row ( (f <fixture>) (rows <parse>) )
  (do-cells f (parts rows)))

(defmethod do-cells ( (f <fixture>) (cells <parse>) )
  (when *debug-fit* (format t "do-cells ~a ~s~%" f cells))
  (iter again ( (cell cells) (column-num 0) )
     (when cell
       (catch-errors 
	cell
	(do-cell f cell column-num))
       (again (more cell) (add1 column-num)))
) )

(defmethod do-cell ( (f <fixture>) (cell <parse>) (column-number integer) )
  (ignore-without-count f cell)
)


(defmethod label-map ( (f <fixture>) (label string) )
"
Given a label from a table, return the name of a function 
which takes a cell and does a set/get/whatever.
"
  label)

(defmethod type-map ( (f <fixture>) (label string) )
"
Given a label from a table, return the type name, a symbol
or T for the unknown/universal type.
"
  t)


;; summary is a simple put-prop/get-prop db
(defmethod summary-put ( (f <fixture>) key value )
  (let ( (bucket (summary-lookup f key)) )
    (if bucket
	(setf (cdr bucket) value)
        (setf (summary f)
	  (cons (cons key value)
		(summary f))))
) )

;; return nil or value
(defmethod summary-get ( (f <fixture>) key )
  (let ( (bucket (summary-lookup f key)) )
    (if bucket (cdr bucket) nil)
) )

;; return nil or (cons key value)
(defmethod summary-lookup ( (f <fixture>) key )
  (labels
      ( (again (buckets)
	   (cond
	    ((null buckets) nil)
	    ((equal key (caar buckets)) (car buckets))
	    (T (again (cdr buckets)))))
      )
    (again (summary f))
) )

(defmethod summary-string ( (f <fixture>) )
  (format nil
	  "~a right, ~a wrong, ~a ignored, ~a exceptions"
	  (num-right f)
	  (num-wrong f)
	  (num-ignores f)
	  (num-excepts f))
)

(defmethod reset-counts ( (f <fixture>) )
"reset class global <fixture> counts"
  (setf (num-right f)    0)
  (setf (num-wrong f)    0)
  (setf (num-ignores f)  0)
  (setf (num-excepts f)  0)
  (setf (summary f) '())
  (summary-put f "run date" (make-current-date))
)

;;==============================================
;; text->value conversions

(defmethod read-from-cell ( (cell <parse>) )
  (with-input-from-string (inp (text cell)) (read inp)))

(defmethod read-number-from-cell ( (cell <parse>) )
  (read-from-cell cell))

;;==============================================
;; Annotations

(defmethod check ( (f <fixture>) (cell <parse>) actual )
  (let* ( (expected (read-from-cell cell)) 
	  (compare  (if (stringp expected) #'string-equal #'equal))
	)
;  (format t "expected=~s, actual=~s~%" expected actual) ;;@@@
    (if (funcall compare expected actual)
	(right f cell)
        (wrong-with-actual f cell (format nil "~s" actual)))))

(defmethod ignore ( (f <fixture>) (cell <parse>) )
  (incf (num-ignores f))
  (ignore-without-count f cell))

(defmethod ignore-without-count ( (f <fixture>) (cell <parse>) )
  (add-to-tag cell +ignore-tag-string+))

(defmethod right ( (f <fixture>) (cell <parse>) )
  (incf (num-right f))
  (add-to-tag cell +right-tag-string+))

(defmethod wrong ( (f <fixture>) (cell <parse>) )
  (incf (num-wrong f))
  (add-to-tag cell +wrong-tag-string+))

(defmethod wrong-with-actual ( (f <fixture>) (cell <parse>) (actual string) )
  (wrong f cell)
  (add-to-body cell
	       (concatenate 'string
			    (label "expected")
			    "<hr>"
			    (escape actual)
			    (label "actual")))
)

;;==============================================
;; Tag text utility functions

(defmethod label ( (str string) )
  (format nil "<font size=-1~a><i> ~a</i></font>" +label-tag-string+ str))

(defmethod gray ( (str string) )
  (format nil "<font~a>~a</font>" +gray-tag-string+ str))

(defmethod escape ( (str string) )
"(ESCAPE string)
replace html used characters with html escape character encoding"
  (let ( (len (length str)) )
    (labels 
	(
	 (add-part (start end escape-str parts)
	    (if (= start end)
		(cons escape-str parts)
		(cons  escape-str 
		      (cons (subseq str start end) parts)))
	 )
	 (gather-parts (parts start end)
	      (apply #'string-append
		   (if (= start end)
		       (reverse parts)
		       (reverse
			(cons (subseq str start end)
			      parts))))
	 )
	 (again (parts start cur)
	    (if (>= cur len)
		(gather-parts parts start cur) ;; done
	        (case (aref str cur)
		  ((#\<) (again (add-part start cur "&lt;" parts)
			       (add1 cur) (add1 cur))
		  )
		  ((#\>) (again (add-part start cur "&gt;" parts)
			       (add1 cur) (add1 cur))
		  )
		  ((#\&) (again (add-part start cur "&amp;" parts)
			       (add1 cur) (add1 cur))
		  )
; 	          ((#\space) (again (add-part start cur "&nbsp;" parts)
; 			      (add1 cur) (add1 cur))
; 	          )
		  ((#\") (again (add-part start cur "&quot;" parts)
			       (add1 cur) (add1 cur))
		  )
		  ;; NYI: unicode, e.g. "&#233;" "&eacute;"
		  (T (again parts start (add1 cur)))))
	 )
	) ;; end labels decls
      (again '() 0 0) ; (again parts start cur)
) ) )


;;;==============================================
;;; class <type-adapter>
;;;==============================================
(defclass <type-adapter> () 
  ;; NB: at least one of SLOT-NAME, METHOD-NAME is unbound
  ( (slot-name   :accessor slot-name   :initarg :slot-name   :initform nil)   ; symbol
    (method-name :accessor method-name :initarg :method-name :initform nil) ; symbol
    (object      :accessor object      :initarg :object      :initform nil)
    (type-sym    :accessor type-sym    :initarg :type        :initform t)
    (fixture     :accessor fixture     :initarg :fixture     :initform nil)
  )
)

(defmethod print-object ( (ta <type-adapter>) port )
   (format port "#[<type-adapter>NAME=~s TYPE=~a]"
	   (get-name ta)
	   (type-sym ta)))

(defmethod parse ( (ta <type-adapter>) (text string) )
  (let ( (desired-type (type-sym ta)) )
   (cond
    ((equal desired-type t) ;; just do the default thing
     (with-input-from-string (inp text) (read inp)))
    ((member desired-type '(float real))
     (+ 0.0 (with-input-from-string (inp text) (read inp))))
    ((member desired-type '(number integer complex rational))
     ;; just do the default thing
     (with-input-from-string (inp text) (read inp)))
    ((equal desired-type 'boolean)
     (let ( (ch (and (> (length text) 0) (aref text 0))) )
       (cond
	((member ch '(#\t #\T)) t)
	((member ch '(#\f #\F)) nil)
	(T (error "Parse: bad boolean ~s" text))))
    )
    ((equal desired-type 'string)
     text)
    ((equal desired-type 'symbol)
     (string->symbol text))
    (T (error "parse: unknown type" desired-type))
) ) )

(defmethod to-string ( (ta <type-adapter>) whatever )
  (let ( (desired-type (type-sym ta)) )
   (cond
    ((equal desired-type 'boolean)
     (if whatever "true" "false")
    )
    (T
     (format nil "~a" whatever))
) ) )

(defmethod get-value ( (ta <type-adapter>) )
  (let ( (value
	  (cond
	   ((slot-name ta) 
	    (slot-value (object ta) (slot-name ta))
	   )
	   ((method-name ta) 
	    (invoke ta)
	   )
	   (T nil)))
       )
    (if (equal 'string (type-sym ta))
	(to-string ta value)
	value)
) )

(defmethod get-name ( (ta <type-adapter>) )
  (cond
   ((slot-name ta) 
    (slot-name ta)
   )
   ((method-name ta) 
    (method-name ta)
   )
   (T nil)
) )

(defmethod set-value ( (ta <type-adapter>) value )
  (setf (slot-value (object ta) (slot-name ta)) value))

(defmethod invoke ( (ta <type-adapter>) )
  (funcall (symbol-function (method-name ta)) (object ta))
)

(defmethod check ( (f <fixture>) (cell <parse>) (adapter <type-adapter>) )
  (let ( (text (text cell)) )
;    (format t "cell text=~s~%" text) ;;@@@
    (cond
     ((equal "" text) (ignore f cell)
     )
     ((equal text "error")
      (block :bogus3121
	(handler-bind
	 ((system::simple-unbound-slot
	   #'(lambda (c) (exception cell c) (return-from :bogus3121 nil))
	   )
	  (condition  
	   #'(lambda (c) 
	       (declare (ignore c))
	       (right f cell) (return-from :bogus3121 nil))
	   )
	  )
	 (invoke adapter)))
      )
     (T (catch-errors cell
	  (let* ( (actual   (get-value adapter))
		  (expected (parse adapter text))
		  (compare  (if (stringp expected) #'string-equal #'equal))
		)
	    (if (funcall compare expected actual)
		(right f cell)
	        (wrong-with-actual f cell (to-string adapter actual))))))
; 				   (format nil "~s parsed as ~s of type ~s" 
; 					   (to-string adapter actual)
; 					   (parse adapter text)
; 					   (type-sym adapter)))))))
) ) )


;;;==============================================
;;; class <column-fixture>
;;;==============================================

(defclass <column-fixture> (<fixture>) 
  ( (column-bindings :accessor column-bindings ) )
)

(defmethod do-rows ( (f <column-fixture>) (rows <parse>) ) 
  (bind f (parts rows))
  (call-next-method f (more rows))
)

(defmethod do-cell ( (f <column-fixture>) (cell <parse>) (column-num integer) )
  (catch-errors cell
    (let ( (binding (nth column-num (column-bindings f)))
	   (text    (text cell))
	 )
;;      (format t "do-cell: col=~s, text=~s, binding=~s~%" column-num text binding) ;;@@@
      (cond
       ((equal "" text) 'ignored)
       ((not binding) (ignore f cell))
       ((slot-name binding) (set-value binding (parse binding text)))
       (T (check f cell binding))
       )
) ) )


(defmethod bind-method ( (f <column-fixture>) (name string) )
  (make-instance '<type-adapter> 
		 :slot-name nil
		 :method-name (string->symbol (label-map f name))
		 :type (type-map f name)
		 :object f
		 :fixture f))

(defmethod bind-slot ( (f <column-fixture>) (name string) )
  (make-instance '<type-adapter> 
		 :slot-name (string->symbol (label-map f name))
		 :type (type-map f name) 
		 :object f
		 :fixture f))

(defmethod bind ( (f <column-fixture>) (cells <parse>) )
  (labels
   ((FIND-SUFFIX (name)
      (let ( (len (length name)) )
	(and (> len 2)
	     (equal "()" (subseq name (- (length name) 2) (length name)))))
    )
    (TRIM-SUFFIX (name) (subseq name 0 (- (length name) 2))
    )
    (TYPE-ADAPTER-FROM-CELL (cell)
       (let ( (name (text cell)) )
	 (catch-errors cell
	   (cond
	    ((equal name "") nil)
	    ((find-suffix name)
	     (bind-method f (trim-suffix name)))
	    (T (bind-slot f name))))
   ) )
   (ITERATE (result cell)
      (if (null cell) 
	  (setf (column-bindings f) (reverse result))
	(iterate (cons (type-adapter-from-cell cell) result) (more cell))))
    )
   (iterate '() cells))
)
	    
 

;;;==============================================
;;; class <action-fixture>
;;;==============================================

(defclass <action-fixture> (<fixture>)
  ( (actor  :accessor actor :initarg :actor
	    :allocation :class :type '<fixture>)
    (cells  :accessor cells :initarg :cells
	    :allocation :class :type '<parse>)
) )

(defmethod do-cells ( (f <action-fixture>) (some-cells <parse>) )
  (setf (cells f) some-cells)
  (catch-errors some-cells
     (let ( (method-sym (string->symbol (string-append "do-" (text some-cells)))) )
;       (format t "do-cells ~a~%" method-sym) ;;@@@
       (funcall (symbol-function method-sym) f)))
)

(defmethod do-start ( (f <action-fixture>) )
  (let* ( (class-sym (string->symbol (text (more (cells f)))))
	  (instance (make-instance class-sym))
	)
    (unless instance
      (error "could not make instance of class" class-sym))
    (setf (actor f) instance)
    (ignore-without-count f (cells f))
) )

(defmethod do-enter ( (f <action-fixture>) )
  (let ( (method-sym (string->symbol
		      (concatenate 'string
				   "enter-"
				   (label-map (actor f) (text (more (cells f)))))))
       )
;    (format t "do-enter: about to call ~a~%" method-sym) ;;@@@
    (funcall (symbol-function method-sym) (actor f) (more (more (cells f))))
    (ignore-without-count f (cells f))
) )

(defmethod do-check ( (f <action-fixture>) )
  (let* ( (method-name (concatenate 'string
				    "check-"
				    (label-map (actor f) (text (more (cells f))))))
	  (method-sym (string->symbol method-name))
       )
;    (format t "do-check: about to call ~a~%" method-sym) ;;@@@
    (let (;; (result (funcall (symbol-function method-sym) (actor f)))
	   (ta (make-instance '<type-adapter> 
			      :method-name method-sym
			      :type (type-map f method-name)
			      :object (actor f)
			      :fixture f))
	)
      (check f (more (more (cells f))) ta)
      (ignore-without-count f (cells f))
) ) )


(defmethod do-press ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "press-"
	    (label-map (actor f) 
		       (text (more (cells f))))))
	  (method-sym (string->symbol method-name))
       )
;     (format t "do-press: about to call ~a~%" method-sym) ;;@@@
      (funcall (symbol-function method-sym) (actor f))
      (ignore-without-count f (cells f))
) )

(defmethod do-await ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "await-"
	    (label-map (actor f) 
		       (text (more (cells f))))))
	  (method-sym (string->symbol method-name))
       )
;   (format t "do-await: about to call ~a~%" method-sym) ;;@@@
    (funcall (symbol-function method-sym) (actor f))
    (ignore-without-count f (cells f))
) )


(defmethod do-pause ( (f <action-fixture>) )
  (let* ( (method-name "pause")
	  (method-sym (string->symbol method-name))
	  (arg-text   (text (more (cells f))))
       )
;   (format t "do-pause: about to call ~a~%" method-sym) ;;@@@
    (let* ( (ta (make-instance '<type-adapter>
		  :method-name method-sym
		  :type (type-map f method-name)
		  :object (actor f)
		  :fixture f))
	    (arg (parse ta arg-text))
	  )
      (funcall (symbol-function method-sym) (actor f) arg)
      (ignore-without-count f (cells f))
) ) )

(defmethod do-fail ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "fail-"
	    (label-map (actor f) 
		       (text (more (cells f))))))
	  (method-sym (string->symbol method-name))
       )
;   (format t "do-fail: about to call ~a~%" method-sym) ;;@@@
    (funcall (symbol-function method-sym) (actor f))
    (ignore-without-count f (cells f))
) )


;;;==============================================
;;; class <timed-action-fixture>
;;;==============================================
(defclass <timed-action-fixture> (<action-fixture>) ())

(defmethod do-table ( (taf <timed-action-fixture>)
		      (table <parse>) )
  (call-next-method taf table)
  (setf (more (last-of (parts (parts table))))
 	(make-instance '<parse> 
		       :start-tag "<td>" :end-tag "</td>" 
		       :body (gray "time")))
   (setf (more (last-of (parts (parts table))))
	 (make-instance '<parse> 
			:start-tag "<td>" :end-tag "</td>" 
			:body (gray "split")))
)

(defmethod do-cells ( (taf <timed-action-fixture>)
		      (cells <parse>) )
  (let ( (start-time    (get-time the-de-simulator)) 
	 (start-seconds (elapsed  the-de-simulator))
       )
    (call-next-method taf cells)
    (let* (;; (end-time    (get-time the-de-simulator))
	    (end-seconds (elapsed  the-de-simulator))
	    (split (+ 0.0
		      (/ (round (* 100 (- end-seconds start-seconds))) 
			 100)))
	  )
      (setf (more (last-of cells))
	    (make-instance '<parse> 
			   :start-tag "<td>" :end-tag "</td>" 
			   :body (gray (date->hms start-time))))
      (setf (more (last-of cells))
	    (make-instance '<parse> 
			   :start-tag "<td>" :end-tag "</td>" 
			   :body (gray (if (<= split 0) 
					   "" 
					 (number->string split)))))
) ) )

;;;==============================================
;;; class <row-fixture>
;;;==============================================
(defclass <row-fixture>  (<column-fixture>) 
  ( (results :accessor results :initarg :results :initform nil)
;;  (missing :accessor missing :initarg :missing :initform '())
    (surplus :accessor surplus :initarg :surplus :initform '())
) )

(defmethod rows->list ( (rows <parse>) )
  (labels
      ( (again (result rows)
	      (if (not rows)
		  (reverse result)
		(again (cons rows result) (more rows))))
      )
  (again '() rows)
) )

(defmethod query ( (f <row-fixture>) )
  (error "query must be implemented by child class of <row-fixture>"))

(defmethod do-rows ( (f <row-fixture>) (rows <parse>) )
  (when *debug-fit* (format t "do-rows: fixture=~a~%" f))
  (bind f (parts rows))
  (setf (results f) (query f))
  (results-missing-surplus f (rows->list (more rows)) (results f))
  (let ( (end (last-of rows)) )
    ;; add surplus rows at end
    (unless (null (surplus f))
	    (setf (more end)
		  (build-rows f (surplus f)))
	    (when *debug-fit*
		  (format t "do-rows: surplus=~a~%" (surplus f))
		  (format t "do-rows: (more end)=~a~%" (more end)))
;;	    (mark f (more end) "surplus")
) ))


(defun match-and-without (obj list)
"MATCH-AND-WITHOUT takes a object and a list
and returns a match or NIL and the list less the match."
  (let ( (found nil) )
    (labels
	  ((key-equal (a b)
	     (equal (car a) (car b)))
	   (without-match (list)
	      (cond
	       ((null list) '())
	       ((key-equal obj (car list))
		(setf found (car list))
		(cdr list))
	       (T
		(cons (car list) (without-match (cdr list)))))
	    )
	  )
	(let ( (residue (without-match list)) )
	  (when *debug-fit* 
	    (format t 
		    "match-and-without: match for ~a is ~a (length residue)=~a~%"
		    obj
		    found
		    (length residue)))
	  (values found residue))
) ) )

;; results-missing-surplus: 
;; common  <- (intersection expected computed)
;; surplus <- (difference computed common)
;; missing <- (difference expected common)
;; done in single pass
(defmethod results-missing-surplus ( (f <row-fixture>)
				     (expected list)
				     (results  list) )
  ;;; NB:
  ;;;   EXPECTED is a list of <parse>
  ;;;   RESULTS  is a list of objects of various types
  (when *debug-fit* (format t "results-missing-surplus ~a~%" f))
  (let* ( (type-adapter (nth 0 (column-bindings f))) 
	  (obj-key (get-name type-adapter))
       )
    (labels
      (
       (package (key val) (cons key val))
       (key     (pkg) (car pkg))
       (val     (pkg) (cdr pkg))
       (cell-key (cell)
		 (parse type-adapter (text (parts cell))))
       (key-of  (obj) 
		(slot-value obj obj-key))
       (wrap-parse (cell)  (package (cell-key cell) cell))
       (wrap-obj   (obj)   (package (key-of   obj)  obj))
      
       ;; WRAP is used to do key extraction once only
       (iterate (expecteds actuals) 
	 (cond
	  ((null expecteds)
	   (setf (surplus f) (mapcar #'val actuals))
	   )
	  ((null actuals)
	   (for-each (lambda (p) (mark f (val p) "missing")) expecteds)
	   )
	  (T
	   (let ( (expected (car expecteds)) )
	     (multiple-value-bind (match dont-match)
		 (match-and-without expected actuals)
	       (if match
		   (check-row f (val expected) (val match))
		 (mark f (val expected) "missing"))
	       (iterate (cdr expecteds) dont-match))))
	) )
       )
      (iterate (mapcar #'wrap-parse expected)
	       (mapcar #'wrap-obj   results))

) ) )


(defmethod check-row ( (f <row-fixture>) (row <parse>) result-object)
  (iter again ( (cell (parts row))
	       (adapters (column-bindings f))
	     ) ;; match bindings
    (unless (null adapters)
       (let ( (type-adapter (car adapters)) )
	 (setf (object type-adapter) result-object)
	 (catch-errors cell
	     (check f cell type-adapter))
	 (again (more cell) (cdr adapters)))
) ) )

(defmethod mark ( (f <row-fixture>) (row <parse>) (message string) )
  (when *debug-fit* 
	(format t "mark ~a ~s~%" row message))
  (unless (and row (parts row))
	  (error (format nil "mark: bad row ~a for msg=~a" row message)))
  (wrong f (parts row))
  (add-to-body (parts row) (label message))
)

(defmethod build-rows ( (f <row-fixture>) (actuals-list list) )
  (when *debug-fit* (format t "build-rows: actuals=~s~%" actuals-list))
  (let ( (root (make-instance '<parse>)) )
    (iter again ( (rows root) (actuals actuals-list) )
      (if (null actuals)
	  (more root) ;; result
	  (let ( (next-row
		  (make-instance '<parse> 
		    :start-tag "<tr>"
		    :parts (build-cells f (car actuals))
		    :end-tag "</tr>"
		  ))
	       )
	    (mark f next-row "surplus")
	    (setf (more rows) next-row)
	    (when *debug-fit*
		  (format t "build-rows: ~s~%" next-row)
		  (print-to-stream next-row)
		  (format t "~%"))
	    (again next-row (cdr actuals))
    ) )  )
) )

(defmethod build-cells ( (f <row-fixture>) surplus-object )
  (when *debug-fit* 
	(format t "build-cells: surplus-obj=~a~%" surplus-object))
  (if (or (not  surplus-object)
	  (null surplus-object))
      ;; if no surplus-object give an empty row
      (make-instance '<parse> 
	    :start-tag "<td>" :end-tag "</td>"
	    :body (format nil "colspan=~a" (length (column-bindings f))))
      ;; else do typical case..
      (let ( (root (make-instance '<parse>)) )
	(iter again ( (cell root) (adapters (column-bindings f)) )
	  (if (null adapters)
	      (more root)
	      (let (  (type-adapter (car adapters))
		      (next-cell (make-instance '<parse> :start-tag "<td>" :body "ack!" :end-tag "</td>"))
		   )
		(setf (more cell) next-cell)
		(catch-errors next-cell
			      (setf (object type-adapter) surplus-object)
			      (setf (body next-cell) 
				    (gray 
				     (escape
				      (to-string type-adapter
						 (get-value type-adapter))))
			      )
	        )
		(when *debug-fit*
		      (format t "build-cells: ~s~%" (body next-cell))
		      (print-to-stream next-cell)
		      (format t "~%"))
		(again next-cell (cdr adapters))
      )) ) )
) )


;;;==============================================
;;; class <summary>
;;;==============================================
(defclass <summary>  (<fixture>) ())

(defmethod do-table ( (f <summary>)
		      (table <parse>) )
  (summary-put f 'counts (summary-string f))
  (setf (more (parts table))
	(make-summary-rows f))
)

(defmethod make-summary-rows ( (f <summary>) )
  (iter again ( (buckets (reverse (summary f)))
	       (result nil)
	     )
    (if (null buckets) 
	result
	(let* ( (bucket (car buckets))
		(key    (car bucket))
		(value  (cdr bucket))
		(new-parse
		 (make-instance '<parse>
		  :start-tag "<tr>" :end-tag "</tr>"
		  :parts (make-instance '<parse> 
			   :start-tag "<td>" :end-tag "</td>"
			   :body (format nil "~a" key)
			   :more (make-instance '<parse> 
				     :start-tag "<td>" :end-tag "</td>"
				     :body (format nil "~a" value)))
		  :more result
	       ))
	      )
	  (when (equal key 'counts)
		(if (or (> (num-wrong   f) 0)
			(> (num-excepts f) 0))
		    (wrong f (more (parts new-parse)))
		    (right f (more (parts new-parse))))
	  )
	  (again (cdr buckets) new-parse)
    )   )
) )

;;; --- E O F --- ;;;
