;;; FILE: "fit.scm"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;
;;;
;;; DATE:    15 October 2002 -- started coding -- KenD
;;; HISTORY: @@@
;;;
;;; NOTES:
;;;  This code parses HTML tables to obtain information on what to test.
;;;  Tables use one of 3 patterns:
;;;	Row     - arg... expected..
;;;	Column  - arg... function
;;;	Actions - action arg... expected...
;;;
;;;  This code parses an input html page, 
;;;	processes the tests,  and 
;;;	outputs a html page of results.
;;;
;;;  The code is not speed critical (fast execution is not a requirement).


;;;==============================================
;;; class <discrete-event-simulator>
;;;=============================================
;;; This discrete event simulator supports three events
;;; each of which is open coded in the body of the simulator.

;;; internal time in absolute seconds
(define-class <discrete-event-simulator> ()
  ( (birth-date :accessor birth-date :type <date>
		:allocation :class
		:init-form (make-current-date))
    (elapsed      :accessor elapsed :type <number> ;; seconds
	       :allocation :class 
	       :init-form 0) ;;(make-current-date))
    (next-search-complete :accessor next-search-complete 
			  :type <number> ;; seconds
			  :allocation :class :init-form 0)
    (next-play-started    :accessor next-play-started 
			  :type <number> ;; seconds
			  :allocation :class :init-form 0)
    (next-play-complete   :accessor next-play-complete 
			  :type <number> ;; seconds
			  :allocation :class :init-form 0)
) )

;; define singleton simulator
(define de-simulator (make <discrete-event-simulator>))

(define (seconds->minutes seconds)
  (/ (round (/ seconds .6)) 100.0))

(define (minutes->seconds minutes)
  (round (* minutes 60)))

(define (hours->seconds hours)
  (* 60 60 hours))


(define-method make ( (ds <discrete-event-simulator>) )
  de-simulator) ;; NB: only 1 instance

(define-method sooner ( (ds <discrete-event-simulator>)
			(soon  <number>)
			(event <number>) )
;; NB: soon must be > elapsed
  (when (< soon (elapsed ds))
	(error "sooner: can't allow negative time"))
  (if (> soon event (elapsed ds)) 
      event
      soon)
)

(define-method next-event ( (ds <discrete-event-simulator>) 
			    (bound <number>) )
;; NB: assert (<= elapsed bound)
  (let ( (next
	  (sooner ds 
		  (sooner ds 
			  (sooner ds
				  bound
				  (next-search-complete ds))
			  (next-play-started ds))
		  (next-play-complete ds)))
       )
    (when *debug-fit*
	  (format #t "next-event: ~a~%" next))
    next
) )

;; send events at specified times
(define-method perform ( (ds <discrete-event-simulator>) )
  (let ( (time (elapsed ds))
	 (search-done (next-search-complete ds))
	 (play-start  (next-play-started    ds))
	 (play-end    (next-play-complete   ds))
       )
    (when (and (> search-done 0)
	       (>= time search-done))
	  (set! (next-search-complete ds) 0)
	  (library-search-complete))
    (when (and (> play-start 0)
	       (>= time play-start))
	  (set! (next-play-started ds) 0)
	  (play-started music-player))
    (when (and (> play-end 0)
	       (>= time play-end))
	  (set! (next-play-complete ds) 0)
	  (play-completed music-player))
) )

(define-method advance ( (ds <discrete-event-simulator>) 
			 (future <number>) )
  (while (< (elapsed ds) future)
	 (when *debug-fit*
	       (format #t 
		       "advance: elapsed=~a future=~a~%" 
		       (elapsed ds) 
		       future)
	       (format #t "advance: next-event=~a~%"
		       (next-event ds future)))
	 (set! (elapsed ds)
	       (next-event ds future)
	 )
	 (perform ds))
)

(define-method schedule ( (ds <discrete-event-simulator>)
			  (seconds <number>) )
"returns minutes from 'now'"
  (+ (elapsed ds) seconds)
)

(define-method delay-for ( (ds <discrete-event-simulator>)
		       (seconds <number>) )
  (advance ds (schedule ds seconds))
)

(define-method wait-search-complete ( (ds <discrete-event-simulator>) )
  (advance ds (next-search-complete ds))
)

(define-method wait-play-started ( (ds <discrete-event-simulator>) )
  (advance ds (next-play-started ds))
)

(define-method wait-play-complete ( (ds <discrete-event-simulator>) )
  (advance ds (next-play-complete ds))
)

(define-method get-time ( (ds <discrete-event-simulator>) )
;; return a date object
  (let* ( (time (elapsed ds))
	  (minutes (truncate (seconds->minutes time)))
	  (seconds (truncate (- time (minutes->seconds minutes))))
	  (delta (make <date> :minutes minutes :seconds seconds))
	)
    (+ delta (birth-date ds))
) )


 
;;;==============================================
;; Color tag info

(define +wrong-tag-string+      " bgcolor=#ffcfcf")
(define +exception-tag-string+  " bgcolor=#ffffcf")
(define +right-tag-string+      " bgcolor=#cfffcf")
(define +ignore-tag-string+     " bgcolor=#efefef")
(define +gray-tag-string+         " color=#808080")
(define +label-tag-string+        " color=#c08080")

;;;==============================================
;;; class <parse>
;;;==============================================

;;; a <parse> instance holds the results of a parse of html text.
(define-class <parse> ()
  ((leader  :init-keyword :leader  :accessor leader  :init-form "") 
   (tag     :init-keyword :tag     :accessor tag) ;; must be specified
   (body    :init-keyword :body    :accessor body    :init-form #f)
   (parts   :init-keyword :parts   :accessor parts   :init-form #f)
   (more    :init-keyword :more    :accessor more    :init-form #f)
   (end     :init-keyword :end     :accessor end) ;; must be specified
   (trailer :init-keyword :trailer :accessor trailer :init-form "")))

(define-method do-parse ( (a-string <string>) (tag-list <list>) )
  (let* ( (tag           (car tag-list))
	  (next-tag-list (cdr tag-list))
	  (start-tag     (string-append "<"  tag))
	  (end-tag       (string-append "</" tag))
	  (start-pos     (search-for start-tag a-string 0))
	  (start-end     (and start-pos (search-for ">"       a-string (add1 start-pos))))
	  (end-pos       (and start-end (search-for end-tag   a-string (add1 start-end))))
	  (end-end       (and end-pos   (search-for ">"       a-string (add1 end-pos))))
	  (next-start    (and end-end   (search-for start-tag a-string (add1 end-end))))
        )
    (when (not end-end)
	  (error "do-parse: can't find tag:" tag))
;      (when *debug-fit* 
;  	  (format #t "start@~a..~a, end@~a..~a, next-start=~a~%"
;  		start-pos start-end  
;  		end-pos end-end
;  		next-start))
    (let* ( (start-end (add1 start-end))
	    (end-end   (add1 end-end))
	    (body    (substring a-string start-end end-pos))
	    (trailer (substring a-string end-end   (string-length a-string)))
	  )
;;      (when *debug-fit* (format #t "body=~s, trailer=~s~%" body trailer))
      (make <parse>
	:leader  (substring a-string 0         start-pos) 
	:tag     (substring a-string start-pos start-end)
	:body    (if (null? next-tag-list)
		     body
		     #f)
	:parts   (if (null? next-tag-list)
		     #f
		     (do-parse body next-tag-list))
	:more    (if next-start
		     (do-parse trailer tag-list)
		     #f)
	:end     (substring a-string end-pos   end-end) 
	:trailer (if (not next-start) trailer ""))
) ) )

(define-method write-object ( (p <parse>) port )
   (format port "#[<parse>TAG=~s]" (tag p)))


(define-method size ( (p <parse>) )
  (if (more p) (add1 (size (more p))) 1))

(define-method last-of ( (p <parse>) )
  (if (more p) (last-of (more p)) p))

(define-method leaf ( (p <parse>) )
  (if (parts p) (leaf (parts p)) p))

(define-method at ( (p <parse>) (i <integer>) )
  (if (or (= i 0) (not (more p))) p (at (more p) (- i 1))))

(define-method at ( (p <parse>) (i <integer>) (j <integer>) )
  (at (parts (at p i)) j))

(define-method at ( (p <parse>) (i <integer>) (j <integer>) (k <integer>) )
  (at (parts (at p i j)) k))

(define-method add-to-tag ( (p <parse>) (text <string>) )
;   (when *debug-fit*
; 	(format #t "add-to-tag: before=~s, substr=~s, " 
; 		(tag p) 
; 		(string-append (substring (tag p) 0 (- (string-length (tag p)) 1)) text ">")))
  (set! (tag p)
	(string-append (substring (tag p) 0 (- (string-length (tag p)) 1)) text ">"))
;;  (when *debug-fit* (format #t "after=~s~%" (tag p)))
)

(define-method add-to-body ( (p <parse>) (text <string>) )
  (set! (body p)
	(if (body p)
	    (string-append (body p) text)
	    text))
)

;;; PRINT-TO-STREAM prints the <parse> P back into HTML
(define-method print-to-stream ( (p <parse>) )
  (print-to-stream p (standard-output-port))) ;; default to stdout

(define-method print-to-stream ( (p <parse>) stream ) 
  (when (leader p) 
	(format stream "~a" (leader p)))
  (format stream "~a" (tag p)) ;; begin tag
  (when (body p) 
	(format stream "~a" (body p)))
  (when (parts p)
	(print-to-stream (parts p) stream))
  (format stream "~a" (end p)) ;; end-tag
  (when (more p)
	(print-to-stream (more p) stream))
  (when (trailer p)
	(format stream "~a" (trailer p)))
  #f
)

(define-method text ( (p <parse>) )
;; returns the string text less tags and whitespece
;   (when *debug-fit*
; 	(unless (body p)
; 		(describe p))) ;;@@@debug
  (if (body p)
      (trim-whitespace (unescape (unformat (body p))))
      "")
)

(define-method exception ( (p <parse>) (error-msg <string>) )
;;  (invoke-debugger c)
  (set! (excepts a-fixture) (add1 (excepts a-fixture)))
  (add-to-body p (format #f "<hr>~a" (escape error-msg)))
  (add-to-tag p +exception-tag-string+))


;;;==============================================
;;; class <fixture>
;;;==============================================
(define-class <fixture> () 
  ( (right   :accessor right    :allocation :class :init-form 0)
    (wrong   :accessor wrong    :allocation :class :init-form 0)
    (ignores :accessor ignores  :allocation :class :init-form 0)
    (excepts :accessor excepts  :allocation :class :init-form 0)
    (summary :accessor summary  :allocation :class :init-form '())
) )

(define a-fixture (make <fixture>)) ;; for (exception <parse> <error>), above.

(define-method do-tables ( (f <fixture>) (tables <parse>) )
  (let loop ( (table tables) )
    (when table
	  (let ( (cell (at table 0 0 0)) )
 	    (when *debug-fit*
 		  (format #t "do-tables: fixture name is ~s~%" 
			  (read-fixture cell)))
	    (catch-errors cell
	     (let* ( (fixture-sym   (read-fixture cell)) 
		     (fixture-class (and fixture-sym (lookup-name fixture-sym)))
		     (fixture	    (and fixture-class (make fixture-class)))
		   )
	       (when *debug-fit*
		     (format #t "do-tables: fixture is ~s~%" fixture))
	       (unless fixture
		       (error 
			(format #f 
				"do-tables: could not make fixture with name ~s" 
				fixture-sym)))
	       (do-table fixture table)))
	    (loop (more table))))
) )

(define-method do-table ( (f <fixture>) (table <parse>) )
  (when *debug-fit* (format #t "do-table ~a~%" table))
  (do-rows f (more (parts table))) ;; skip fixture name
)

(define-method read-fixture ( (cell <parse>) )
; reads a class-name from the 1st cell in a parse instance
; or throws an error;
; returns a symbol.
  (string->symbol (caseify (text cell)))
)

(define-method do-rows ( (f <fixture>) (rows <parse>) )
  (let loop ( (row rows) )
    (when row
       ;; be careful of rows added by processing
       (let ( (next (more row)) )
	 (do-row f row)
	 (loop next)))
) )


(define-method do-row ( (f <fixture>) (rows <parse>) )
  (when *debug-fit* (format #t "do-row ~a~%" rows))
  (do-cells f (parts rows)))

(define-method do-cells ( (f <fixture>) (cells <parse>) )
  (when *debug-fit* (format #t "do-cells ~a ~s~%" f cells))
  (let loop ( (cell cells) (column-num 0) )
    (when cell
	  (catch-errors 
	    cell
	    (do-cell f cell column-num))
	  (loop (more cell) (add1 column-num)))
) )

(define-method do-cell ( (f <fixture>) (cell <parse>) (column-number <integer>) )
  (when *debug-fit* (format #t "do-cell ~a~%" cell))
  (ignore-without-count f cell)
)


(define-method label-map ( (f <fixture>) (label <string>) )
; Given a label from a table, return the name of a function 
; which takes a cell and does a set/get/whatever.
  label)

(define-method type-map ( (f <fixture>) (label <string>) )
; Given a label from a table, return the type name, a symbol
; or #T for the unknown/universal type.
  #t)


;; summary is a simple put-prop/get-prop db
(define-method summary-put ( (f <fixture>) key value )
  (cond
   ((summary-lookup f key) 
     => (lambda (bucket)
	  (set! (cdr bucket) value))
   )
   (else
    (set! (summary f)
	  (cons (cons key value)
		(summary f))))
) )

;; return #f or value
(define-method summary-get ( (f <fixture>) key )
  (cond
   ((summary-lookup f key) => (lambda (bucket) (cdr bucket)))
   (else #f)
) )

;; return #f or (cons key value)
(define-method summary-lookup ( (f <fixture>) key )
  (let loop ( (buckets (summary f)) )
    (cond
     ((null? buckets) #f)
     ((equal? key (caar buckets)) (car buckets))
     (else (loop (cdr buckets)))
) ) )

(define-method summary-string ( (f <fixture>) )
  (format #f 
	  "~a right, ~a wrong, ~a ignored, ~a exceptions"
	  (right f)
	  (wrong f)
	  (ignores f)
	  (excepts f))
)

(define-method reset-counts ( (f <fixture>) )
  (set! (right f)    0)
  (set! (wrong f)    0)
  (set! (ignores f)  0)
  (set! (excepts f)  0)
  (set! (summary f) '())
  (summary-put f "run date" (make-current-date))
)


;;==============================================
;; text->value conversions

(define-method read-from-cell ( (cell <parse>) )
  (call-with-input-string (text cell)
			  (lambda (p) (read p)))
)

(define-method read-number-from-cell ( (cell <parse>) )
  (read-from-cell cell))

;;==============================================
;; Annotations

(define-method check ( (f <fixture>) (cell <parse>) actual )
  (let ( (expected (read-from-cell cell)) )
    (when *debug-fit* 
	  (format #t "expected=~s, actual=~s~%" expected actual))
    (if (if (and (number? expected) (number? actual))
	    (= expected actual)
	    (equal? expected actual))
	(right f cell)
        (wrong-with-actual f cell (format #f "~s" actual)))))

(define-method ignore ( (f <fixture>) (cell <parse>) )
  (set! (ignores f) (add1 (ignores f)))
  (add-to-tag cell +ignore-tag-string+))

(define-method ignore-without-count ( (f <fixture>) (cell <parse>) )
  (add-to-tag cell +ignore-tag-string+))

(define-method right ( (f <fixture>) (cell <parse>) )
  (set! (right f) (add1 (right f)))
  (add-to-tag cell +right-tag-string+))

(define-method wrong ( (f <fixture>) (cell <parse>) )
  (set! (wrong f) (add1 (wrong f)))
  (add-to-tag cell +wrong-tag-string+))

(define-method wrong-with-actual ( (f <fixture>) (cell <parse>) (actual <string>) )
  (wrong f cell)
  (add-to-body cell
	       (string-append (label "expected")
			      "<hr>"
			      (escape actual)
			      (label "actual")))
)

;;==============================================
;; Tag text utility functions

(define-method label ( (str <string>) )
  (format #f "<font size=-1~a><i> ~a</i></font>" +label-tag-string+ str))

(define-method gray ( (str <string>) )
  (format #f "<font~a>~a</font>" +gray-tag-string+ str))

(define-method escape ( (str <string>) )
;replace html used characters with html escape character encoding
  (let ( (len (string-length str))
	 (add-part
	  (lambda (start end escape-str parts)
	    (if (= start end)
		(cons escape-str parts)
		(cons  escape-str 
		      (cons (substring str start end) parts)))))
	 (gather-parts
	  (lambda (parts start end)
	    (apply string-append
		   (if (= start end)
		       (reverse parts)
		       (reverse
			(cons (substring str start end)
			      parts))))))
       )
    (let loop ( (parts '()) (start 0) (cur 0) )
      (if (>= cur len)
	  (gather-parts parts start cur) ;; done
	  (case (string-ref str cur)
	    ((#\<) (loop (add-part start cur "&lt;" parts)
			 (add1 cur) (add1 cur))
	    )
	    ((#\>) (loop (add-part start cur "&gt;" parts)
			 (add1 cur) (add1 cur))
	    )
	    ((#\&) (loop (add-part start cur "&amp;" parts)
			 (add1 cur) (add1 cur))
	    )
; 	    ((#\space) (loop (add-part start cur "&nbsp;" parts)
; 			 (add1 cur) (add1 cur))
; 	    )
	    ((#\") (loop (add-part start cur "&quot;" parts)
			 (add1 cur) (add1 cur))
	    )
	    ;; unicode, e.g. "&#233;" "&eacute;"
	    (else (loop parts start (add1 cur))))
) ) ) )


(define-method unformat ( (str <string>) )
;; remove <> tags
  (letrec
      ((iter 
	(lambda (str)
	  (let ( (more (search-for "<" str 0)) )
	    (if more
		(let ( (tag-start more)
		       (tag-end (search-for ">" str (+ more 1)))
		     )
		  (if tag-end
		      (iter (string-append (substring str 0 tag-start)
					   (substring str (+ 1 tag-end) (string-length str))))
		      str))
		str)
	)))
       )
    (iter str)
) )

(define-method unescape ( (str <string>) )
;; convert html escapes into text (e.g. &lt; -> <)
  (letrec
      ((subst 
	(lambda (esc)
;	 (format t "subst for ~s~%" esc) ;;@@@
	 (cond
	  ((equal? esc "&lt;")   "<")
	  ((equal? esc "&gt;")   ">")
	  ((equal? esc "&amp;")  "&")
	  ((equal? esc "&nbsp;") " ")
	  ((equal? esc "&quot;") "\"")
	  ;; unicode, e.g. "&#233;" "&eacute;"
	  (else #f))
       ))
       (iter 
	(lambda (str pos)
	  (let ( (more (search-for "&" str pos)) )
	    (if more
		(let* ( (esc-start more)
		        (esc-end (and more (search-for ";" str (add1 esc-start))))
			(len (string-length str))
			(double&? (and more 
				       (< (add1 esc-start) len)
				       (char=? (string-ref str (add1 esc-start)) #\&)))
		     )
;		  (format t "escape from=~a to=~a~%" esc-start esc-end) ;;@@@
		  (cond
		   (double&? ;; "&&" => skip first "&"
		    (iter str (add1 esc-start))
		   )
		   (esc-end
		    (let* ( (subst-piece (or (subst 
					      (substring str 
							 esc-start 
							 (+ esc-end 1)))
					     ""))
			    (next-pos (+ esc-start (string-length subst-piece)))
			    )
		      (iter (string-append (substring str 0 esc-start)
					   subst-piece
					   (substring str (+ 1 esc-end) (string-length str)))
			    next-pos))
		   )
		   (else str)))
		str)
        )))
       )
       (iter str 0)
) )

(define-method trim-whitespace ( (str <string>) )
;; return string less leading and trailing whitespace
  (let ( (start 0)
	 (len (string-length str))
       )
    (letrec
	((get-start 
	  (lambda (cur)
	    (cond
	     ((>= cur len) "")
	     ((char-whitespace? (string-ref str cur))
	      (set! start (add1 cur))
	      (get-start (add1 cur)))
	     (else (get-end (sub1 len))))
	  ))
	 (get-end
	  (lambda (cur)
	    (cond
	     ((< cur 0) (substring str start (+ cur 1)))
	     ((char-whitespace? (string-ref str cur))
	      (get-end (sub1 cur)))
	     (else (substring str start (+ cur 1))))
	  ))
	 )
      (get-start 0)
) ) )


;;;==============================================
;;; class <type-adapter>
;;;==============================================

(define-class <type-adapter> () 
  ;; NB: at least one of SLOT-NAME, METHOD-NAME is unbound
  ( (slot-name   :accessor slot-name   
		 :init-keyword :slot-name   
		 :init-form #f) ; symbol
    (method-name :accessor method-name 
		 :init-keyword :method-name 
		 :init-form #f) ; symbol
    (object      :accessor object      
		 :init-keyword :object      
		 :init-form #f)
    (type-sym    :accessor type-sym    
		 :init-keyword :type        
		 :init-form #t) ; #t => universal type
    (fixture     :accessor fixture     
		 :init-keyword :fixture     
		 :init-form #f)
  )
)

(define-method write-object ( (ta <type-adapter>) port )
   (format port "#[<type-adapter>NAME=~s TYPE=~a]"
	   (get-name ta)
	   (type-sym ta)))


(define-method parse ( (ta <type-adapter>) (text <string>) )
  (let ( (desired-type (type-sym ta)) )
    (when *debug-fit*
	  (format #t "parse: type=~a, text=~s~%" desired-type text)
	  (format #t "       fixture=~a, adapter=~a~%" (fixture ta) ta)
	  (format #t "       method-name=~a~%" (method-name ta)))
   (cond
    ((equal? desired-type #t) ;; just do the default thing
     (call-with-input-string text (lambda (p) (read p)))
    )
    ((memq desired-type '(float real))
     (+ 0.0 (call-with-input-string text (lambda (p) (read p))))
    )
    ((memq desired-type '(number integer complex rational))
     (call-with-input-string text (lambda (p) (read p))) ;; just use READ
    )
    ((equal? desired-type 'boolean)
     (let ( (ch (and (> (string-length text) 0) (string-ref text 0))) )
       (cond
	((member ch '(#\t #\T)) #t)
	((member ch '(#\f #\F)) #f)
	(else (error "Parse: bad boolean ~s" text))))
    )
    ((equal? desired-type 'string)
     text
     )
    ((equal? desired-type 'symbol)
     (string->symbol (caseify text))
    )
    (else (error (format #f "parse: unknown type ~s" desired-type)))
) ) )

(define-method to-string ( (ta <type-adapter>) whatever )
  (let ( (desired-type (type-sym ta)) )
   (cond
    ((equal? desired-type 'boolean)
     (if whatever "true" "false")
    )
     ((memq desired-type '(number integer real rational float complex))
      (if (and (number? whatever) (inexact? whatever))
 	 (format #f "inexact: ~a" whatever)
 	 (format #f "~a" whatever))
     )
    (else
     (format #f "~a" whatever))
) ) )

(define-method get-value ( (ta <type-adapter>) )
  (let ( (value
	  (cond
	   ((slot-name ta) 
	    (slot-value (object ta) (slot-name ta))
	    )
	   ((method-name ta) 
	    (invoke ta)
	    )
	   (else #f)))
       )
    (if (equal? 'string (type-sym ta))
	(to-string ta value)
	value)
) )

;; for write-object
(define-method get-name ( (ta <type-adapter>) )
  (cond
   ((slot-name ta) => (lambda (x) x)
   )
   ((method-name ta) => (lambda (x) x)
   )
   (else #f)
) )

(define-method set-value ( (ta <type-adapter>) value )
  (set! (slot-value (object ta) (slot-name ta)) value))

(define-method invoke ( (ta <type-adapter>) )
  (when *debug-fit*
	(format #t "invoking (~s ~s)~%" (method-name ta) (object ta)))
  ;; apply method by name to the type-adapter's object
  (let ( (method (lookup-name (method-name ta))) )
    (if method
	(method (object ta))
	(error (format #f 
		       "invoke: could not find method named ~s" 
		       (method-name ta))))
) )


(define-method check ( (f <fixture>) (cell <parse>) (adapter <type-adapter>) )
  (let ( (text (text cell)) )
    (when *debug-fit* 
	  (format #t "cell text=~s~%" text))
    (cond
     ((equal? "" text) (ignore f cell)
     )
     ((string-ci=? text "error")
      ;;(catch-errors cell (invoke adapter))
      (right f cell)
      )
     (else (catch-errors cell
	  (let* ( (actual (get-value adapter)) 
		  (type   (type-sym  adapter))
		  (is-equal? (if (memq type '(number integer real rational float complex)) 
				 = 
				 equal?)) 
		)
;;	    (format #t "type=~a, is-equal?=~a~%" type is-equal?) ;;@@@
	    (if (is-equal? (parse adapter text) actual)
		(right f cell)
		(begin
		  (wrong-with-actual f cell (to-string adapter actual))
		  (when *debug-fit*
			(format #t 
				"expected ~s, got ~s; type ~s using compare ~s~%" 
				(parse adapter text)
				actual ;(to-string adapter actual)
				(type-sym adapter)
				is-equal?))
		  ))))))
) )


;;;==============================================
;;; class <column-fixture>
;;;==============================================

(define-class <column-fixture> (<fixture>) 
  ( (column-bindings :accessor column-bindings ) )
)

(define-method do-rows ( (f <column-fixture>) (rows <parse>) ) 
  (bind f (parts rows))
  (let loop ( (row (more rows)) )
    (when row
    ;; be careful of rows added by processing
      (let ( (next (more row)) )
	(do-row f row)
	(loop next)))
) )

(define-method do-cell ( (f <column-fixture>) (cell <parse>) (column-num <integer>) )
  (when *debug-fit*
	(format #t "do-cell: doing ~s number ~a" cell column-num))
  (catch-errors cell
    (let ( (binding (nth column-num (column-bindings f)))
	   (text    (text cell))
	 )
      (when *debug-fit*
	    (format #t "with text=~s and binding=~s~%" text binding))
      (cond
       ((equal? "" text) 'ignored)
       ((not binding) (ignore f cell))
       ((slot-name binding) (set-value binding (parse binding text)))
       (else (check f cell binding))
       )
) ) )


(define-method bind-method ( (cf <column-fixture>) (name <string>) )
  (make <type-adapter> 
    :slot-name #f
    :method-name (string->symbol (caseify (label-map cf name)))
    :type (type-map cf name)
    :object  cf
    :fixture cf))

(define-method bind-slot ( (cf <column-fixture>) (name <string>) )
  (make <type-adapter> 
    :slot-name (string->symbol (caseify (label-map cf name)))
    :method-name #f
    :type (type-map cf name) 
    :object  cf
    :fixture cf))

(define-method bind ( (f <column-fixture>) (cells <parse>) )
  (letrec
   ((find-suffix
     (lambda (name)
       (let ( (len (string-length name)) )
	 (and (> len 2)
	      (equal? "()" 
		      (substring name 
				 (- (string-length name) 2) 
				 (string-length name)))))
    ))
    (trim-suffix 
     (lambda (name) (substring name 0 (- (string-length name) 2)))
    )
    (type-adapter-from-cell
     (lambda (cell)
       (let ( (name (text cell)) )
	 (catch-errors cell
	   (cond
	    ((equal? name "") #f)
	    ((find-suffix name)
	     (bind-method f (trim-suffix name)))
	    (else (bind-slot f name))))
   )) )
   (iter 
    (lambda (result cell)
      (if (not cell) 
	  (set! (column-bindings f) (reverse result))
	(iter (cons (type-adapter-from-cell cell) result) (more cell))))
    )
   )
   (iter '() cells))
)
	    
;;;==============================================
;;; class action-fixture
;;;==============================================

(define-class <action-fixture> (<fixture>)
  ( (actor-of  :accessor actor-of 
	       :init-keyword :actor 
	       :allocation :class 
	       :init-form 'actor-not-set 
	       :type <fixture>)
    (cells-of  :accessor cells-of 
	       :init-keyword :cells 
	       :init-form 'cells-not-set 
	       :type <parse>)
) )

(define-method do-cells ( (f <action-fixture>) (cells <parse>) )
  (when *debug-fit*
	(format #t 
		"do-cells: fixture ~a doing cell ~a with text=~s~%" 
		f 
		cells 
		(text cells)))
  (set! (cells-of f) cells)
  (catch-errors cells
     (let* ( (method-sym (string->symbol 
			  (caseify (string-append "do-" (text cells)))))
	     (method     (lookup-name method-sym))
	   )
       (when *debug-fit*
	     (format #t 
		     "do-cells: about to call method named ~a~%"
		     method-sym))
       (if method
	   (method f)
	   (error (format #f 
			  "do-cells: could not find method for ~a" 
			  method-sym)))
) ) )

(define-method do-start ( (f <action-fixture>) )
  (let* ( (class-sym (string->symbol 
		      (caseify (text (more (cells-of f))))))
	  (class     (lookup-name class-sym))
	  (instance  (and class (make class)))
	)
    (unless instance
      (error "could not make instance of class" class-sym))
    (set! (actor-of f) instance)
    (ignore-without-count f (cells-of f))
) )

(define-method do-enter ( (f <action-fixture>) )
  (let* ( (method-sym 
	   (string->symbol
	    (caseify
	     (string-append 
	      "enter-"
	      (label-map (actor-of f)
			 (text (more (cells-of f))))))))
	  (method (lookup-name method-sym))
        )
    (when *debug-fit*
	  (format #t "do-enter: about to call ~a~%" method-sym))
    (if method
	(begin
	  (method (actor-of f) (more (more (cells-of f))))
	  (ignore-without-count f (cells-of f)))
	(error (format #f 
		       "do-enter: could not find method for ~a" 
		       method-sym)))
) )

(define-method do-check ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "check-"
	    (label-map (actor-of f) 
		       (text (more (cells-of f))))))
	  (method-sym (string->symbol (caseify method-name)))
	  (method     (lookup-name method-sym))
       )
;   (format t "do-check: about to call ~a~%" method-sym) ;;@@@
    (unless method
	    (error (format #f 
			   "do-check: could not find method for ~a" 
			   method-sym)))
    (let ( (ta (make <type-adapter>
		  :method-name method-sym
		  :type (type-map f method-name)
		  :object (actor-of f)
		  :fixture f))
	  )
      (check f (more (more (cells-of f))) ta)
      (ignore-without-count f (cells-of f))
) ) )

(define-method do-press ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "press-"
	    (label-map (actor-of f) 
		       (text (more (cells-of f))))))
	  (method-sym (string->symbol (caseify method-name)))
	  (method     (lookup-name method-sym))
       )
;   (format t "do-press: about to call ~a~%" method-sym) ;;@@@
    (unless method
	    (error (format #f 
			   "do-press: could not find method for ~a" 
			   method-sym)))
    (method (actor-of f))
    (ignore-without-count f (cells-of f))
) )

(define-method do-await ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "await-"
	    (label-map (actor-of f) 
		       (text (more (cells-of f))))))
	  (method-sym (string->symbol (caseify method-name)))
	  (method     (lookup-name method-sym))
       )
;   (format t "do-await: about to call ~a~%" method-sym) ;;@@@
    (unless method
	    (error (format #f 
			   "do-await: could not find method for ~a" 
			   method-sym)))
    (method (actor-of f))
    (ignore-without-count f (cells-of f))
) )


(define-method do-pause ( (f <action-fixture>) )
  (let* ( (method-name "pause")
	  (method-sym (string->symbol method-name))
	  (method     (lookup-name method-sym))
	  (arg-text   (text (more (cells-of f))))
       )
;   (format t "do-pause: about to call ~a~%" method-sym) ;;@@@
    (unless method
	    (error (format #f 
			   "do-pause: could not find method for ~a" 
			   method-sym)))
    (let* ( (ta (make <type-adapter>
		  :method-name method-sym
		  :type (type-map f method-name)
		  :object (actor-of f)
		  :fixture f))
	    (arg (parse ta arg-text))
	  )
      (method (actor-of f) arg)
      (ignore-without-count f (cells-of f))
) ) )

(define-method do-fail ( (f <action-fixture>) )
  (let* ( (method-name 
	   (string-append
	    "fail-"
	    (label-map (actor-of f) 
		       (text (more (cells-of f))))))
	  (method-sym (string->symbol (caseify method-name)))
	  (method     (lookup-name method-sym))
       )
;   (format t "do-fail: about to call ~a~%" method-sym) ;;@@@
    (unless method
	    (error (format #f 
			   "do-fail: could not find method for ~a" 
			   method-sym)))
    (let ( (ta (make <type-adapter>
		  :method-name method-sym
		  :type (type-map f method-name)
		  :object (actor-of f)
		  :fixture f))
         )
      (invoke ta)
      (ignore-without-count f (cells-of f))
) ) )


;;;==============================================
;;; class <timed-action-fixture>
;;;==============================================
(define-class <timed-action-fixture> (<action-fixture>) ())

(define-method do-table ( (taf <timed-action-fixture>)
			  (table <parse>) )
  (next-method)
   (set! (more (last-of (parts (parts table))))
 	(make <parse> :tag "<td>" :end "</td>" 
 	      :body (gray "time")))
   (set! (more (last-of (parts (parts table))))
 	(make <parse> :tag "<td>" :end "</td>" 
 	      :body (gray "split")))
)

(define-method do-cells ( (taf <timed-action-fixture>)
			  (cells <parse>) )
  (let ( (start-time    (get-time de-simulator)) 
	 (start-seconds (elapsed de-simulator))
       )
    (next-method)
    (let* ( (end-time (get-time de-simulator))
	    (end-seconds (elapsed de-simulator))
	    (split (/ (round (* 100 (- end-seconds start-seconds))) 
		      100))
	  )
      (set! (more (last-of cells))
	    (make <parse> :tag "<td>" :end "</td>" 
		  :body (gray (date->hms start-time))))
      (set! (more (last-of cells))
	    (make <parse> :tag "<td>" :end "</td>" 
		  :body (gray (if (<= split 0) "" (number->string split)))))
) ) )

;;;==============================================
;;; class <row-fixture>
;;;==============================================
(define-class <row-fixture>  (<column-fixture>) 
  ( (results :accessor results :init-keyword :results :init-form #f)
;;  (missing :accessor missing :init-keyword :missing :init-form '())
    (surplus :accessor surplus :init-keyword :surplus :init-form '())
) )

(define-method rows->list ( (rows <parse>) )
  (let loop ( (result '()) (rows rows) )
    (if (not rows)
	(reverse result)
	(loop (cons rows result) (more rows))
) ) )

(define-method query ( (f <row-fixture>) )
  (error "query must be implemented by child class of <row-fixture>"))

(define-method do-rows ( (f <row-fixture>) (rows <parse>) )
  (when *debug-fit* (format #t "do-rows: fixture=~a~%" f))
  (bind f (parts rows))
  (set! (results f) (query f))
  (results-missing-surplus f (rows->list (more rows)) (results f))
  (let ( (end (last-of rows)) )
    ;; add surplus rows at end
    (unless (null? (surplus f))
	    (set! (more end)
		  (build-rows f (surplus f)))
	    (when *debug-fit*
		  (format #t "do-rows: surplus=~a~%" (surplus f))
		  (format #t "do-rows: (more end)=~a~%" (more end)))
;;	    (mark f (more end) "surplus")
) ))

;; return a function which takes a object and a list
;; and returns a match or #f and the list less the match.
(define (make-match-and-without same?)
  (lambda (obj list)
    (let ( (found #f) )
      (define (without-match list)
	(cond
	 ((null? list) '())
	 ((same? obj (car list))
	  (set! found (car list))
	  (cdr list))
	 (else
	  (cons (car list) (without-match (cdr list)))))
	)
      (let ( (residue (without-match list)) )
	(when *debug-fit* 
	      (format #t 
		      "match-and-without: match for ~a is ~a (length residue)=~a~%"
		      obj
		      found
		      (length residue)))
	(values found residue))
) ) )

;; results-missing-surplus: 
;; common  <- (intersection expected computed)
;; surplus <- (difference computed common)
;; missing <- (difference expected common)
;; done in single pass
(define-method results-missing-surplus ( (f <row-fixture>)
					 (expected <list>)
					 (results  <list>) )
  ;;; NB:
  ;;;   expected is a list of <parse>
  ;;;   results  is a list of objects of various types
  (when *debug-fit* (format #t "results-missing-surplus ~a~%" f))
  (let* ( (type-adapter (nth 0 (column-bindings f))) 
	  (package   cons) ;(lambda (key val) (cons key val)))
	  (key       car)  ;(lambda (pkg) (car pkg)))
	  (unpackage cdr)  ;(lambda (pkg) (cdr pkg)))
	  (cell-key (lambda (cell)
; 		     (when *debug-fit* 
; 			   (format #t "cell-key ~a ~a~%" cell type-adapter))
; 		     (when *debug-fit* 
; 			   (format #t "         text=~s~%" (text (parts cell))))
; 		     (when *debug-fit* 
; 			   (format #t "         key=~s~%" (parse type-adapter (text (parts cell)))))
		     (parse type-adapter (text (parts cell)))))
	  (the-obj-key (get-name type-adapter))
	  (obj-key  (lambda (obj) 
; 		     (when *debug-fit* 
; 			   (format #t "obj-key ~a is ~s~%" obj (slot-value obj the-obj-key)))
		     (slot-value obj the-obj-key)))
	  (wrap-parse (lambda (cell)  (package (cell-key cell) cell)))
	  (wrap-obj   (lambda (obj)   (package (obj-key   obj)  obj)))
	  (compare? ;; true iff keys are equal
	   (lambda (one another) 
	     (let ( (one-key     (key one))
		    (another-key (key another))
		  )
	       (if (and (number? one-key) (number? (another-key)))
		   (=      one-key another-key)
		   (equal? one-key another-key)))
	  ))
	  ;; takes obj & list, returns values: match-or-#f dont-match
	  (match-and-without (make-match-and-without compare?))
	)
    ;; WRAP is used to do key extraction once only
    (let loop ( (expecteds (map wrap-parse expected))
		(actuals   (map wrap-obj   results))
	      )
      (cond
       ((null? expecteds)
	(set! (surplus f) (map unpackage actuals))
       )
       ((null? actuals)
	(for-each (lambda (p) (mark f (unpackage p) "missing")) expecteds)
       )
       (else
	(let ( (expected (car expecteds)) )
	  (receive (match dont-match)
	      ;; producer
	      (match-and-without expected actuals)
	      ;; use
	      (if match
		  (check-row f (unpackage expected) (unpackage match))
		  (mark f (unpackage expected) "missing")
	      )
	      (loop (cdr expecteds) dont-match))
     )))
) ) )


(define-method check-row ( (f <row-fixture>) (row <parse>) result-object)
  (let loop ( (cell (parts row))
	      (adapters (column-bindings f))
	    ) ;; match bindings
    (unless (null? adapters)
       (let ( (type-adapter (car adapters)) )
	 (set! (object type-adapter) result-object)
	 (catch-errors cell
	     (check f cell type-adapter))
	 (loop (more cell) (cdr adapters)))
) ) )

(define-method mark ( (f <row-fixture>) (row <parse>) (message <string>) )
  (when *debug-fit* 
	(format #t "mark ~a ~s~%" row message))
  (unless (and row (parts row))
	  (error (format #f "mark: bad row ~a for msg=~a" row message)))
  (wrong f (parts row))
  (add-to-body (parts row) (label message))
)

(define-method build-rows ( (f <row-fixture>) (actuals-list <list>) )
  (when *debug-fit* (format #t "build-rows: actuals=~s~%" actuals-list))
  (let ( (root (make <parse>)) )
    (let loop ( (rows root) (actuals actuals-list) )
      (if (null? actuals)
	  (more root) ;; result
	  (let ( (next-row
		  (make <parse> 
		    :tag "<tr>"
		    :parts (build-cells f (car actuals))
		    :end "</tr>"
		  ))
	       )
	    (mark f next-row "surplus")
	    (set! (more rows) next-row)
	    (when *debug-fit*
		  (format #t "build-rows: ~s~%" next-row)
		  (print-to-stream next-row)
		  (format #t "~%"))
	    (loop next-row (cdr actuals))
    ) )  )
) )

(define-method build-cells ( (f <row-fixture>) surplus-object )
  (when *debug-fit* 
	(format #t "build-cells: surplus-obj=~a~%" surplus-object))
  (if (or (not   surplus-object)
	  (null? surplus-object))
      ;; if no surplus-object give an empty row
      (make <parse> 
	    :tag "<td>" :end "</td>"
	    :body (format #f "colspan=~a" (length (column-bindings f))))
      ;; else do typical case..
      (let ( (root (make <parse>)) )
	(let loop ( (cell root) (adapters (column-bindings f)) )
	  (if (null? adapters)
	      (more root)
	      (let (  (type-adapter (car adapters))
		      (next-cell (make <parse> :tag "<td>" :body "ack!" :end "</td>"))
		   )
		(set! (more cell) next-cell)
		(catch-errors next-cell
			      (set! (object type-adapter) surplus-object)
			      (set! (body next-cell) 
				    (gray 
				     (escape
				      (to-string type-adapter
						 (get-value type-adapter))))
			      )
	        )
		(when *debug-fit*
		      (format #t "build-cells: ~s~%" (body next-cell))
		      (print-to-stream next-cell)
		      (format #t "~%"))
		(loop next-cell (cdr adapters))
      )) ) )
) )


;;;==============================================
;;; class <summary>
;;;==============================================
(define-class <summary>  (<fixture>) ())

(define-method do-table ( (f <summary>)
			  (table <parse>) )
  (summary-put f 'counts (summary-string f))
  (set! (more (parts table))
	(make-summary-rows f))
)

(define-method make-summary-rows ( (f <summary>) )
  (let loop ( (buckets (reverse (summary f)))
	      (result #f)
	    )
    (if (null? buckets) 
	result
	(let* ( (bucket (car buckets))
		(key    (car bucket))
		(value  (cdr bucket))
		(new-parse
		 (make <parse>
		  :tag "<tr>" :end "</tr>"
		  :parts (make <parse> 
			   :tag "<td>" :end "</td>"
			   :body (format #f "~a" key)
			   :more (make <parse> 
				     :tag "<td>" :end "</td>"
				     :body (format #f "~a" value)))
		  :more result
	       ))
	      )
	  (when (equal? key 'counts)
		(if (or (> (wrong   f) 0)
			(> (excepts f) 0))
		    (wrong f (more (parts new-parse)))
		    (right f (more (parts new-parse))))
	  )
	  (loop (cdr buckets) new-parse)
    )   )
) )

 
;;; --- E O F --- ;;;
