;;; FILE: "glue.scm"
;;; IMPLEMENTS: Framework for Integrated Test
;;; AUTHOR: Ken Dickey & Ward Cunningham
;;;  See http://fit.c2.com for documentation and examples

;;;
;;; Copyright (c) 2002 Cunningham & Cunningham, Inc.
;;; Released under the terms of the GNU General Public License version 2 or later.
;;;

(define *debug-fit* #f)


(define (add1 n) (+ n 1))
(define (sub1 n) (- n 1))

(define (nth n list)
  (if (<= n 0) 
      (car list) 
      (nth (sub1 n) (cdr list)))
)

; srfi-8
(define-syntax receive
  (syntax-rules ()
    ((receive ?vars ?producer ?body ...) ;=>
     (call-with-values 
	 (lambda () ?producer)
         (lambda ?vars ?body ...))
) ) )

(define caseify 
  ; return string is canonnical intern case
  (let ( (case-fun (if (char=? #\a (string-ref (symbol->string 'A) 0))
		       char-downcase
		       char-upcase))
       )
    (lambda (string)
      (list->string (map case-fun (string->list string))))
) )

(define (newline? char)
  (memq char '(#\newline #\return)))

(define (read-line inport)
  (let loop ( (ch (read-char inport)) (chars '()) )
    (cond
     ((eof-object? ch)
      (if (< 0 (length chars))
	  (list->string (reverse chars))
	  ch))
     ((newline? ch)
      (list->string (reverse chars)))
     (else
      (loop (read-char inport) (cons ch chars)))
) ) )

(define (file->string file-name-string)
  (call-with-input-file 
     file-name-string
     (lambda (inport) (port->string inport))))

;; simple string search (kept to parallel CommonLisp code)
(define (search-for want to-search start-pos)
;;  (when *debug-fit* 
;;	(format #t "(search-for ~s ~s ~a)~%" want to-search start-pos))
  (let ( (want-len      (string-length want))
	 (to-search-len (string-length to-search))
       )
    (letrec
	((try 
	  (lambda (pos)
	    (let ( (part (substring to-search pos (+ pos want-len))) )
		;;(when *debug-fit* (format #t "offset: ~a substring: \"~a\"~%" pos part))
		(string-ci=? part want)) ;; NB: Case Insensitive (ci)
	 ))
	 (iter 
	  (lambda (pos)
	   (cond
	    ((> (+ pos want-len) to-search-len) #f)
	    ((try pos) pos)
	    (else (iter (add1 pos))))
	))
       )
      (iter start-pos))))

(define-syntax catch-errors
  (syntax-rules ()
    ((catch-errors ?cell ?form ...) ;=>
     (with-error-handler
      (lambda (condition) ;; add exception text to cell
	(when *debug-fit*
	      (format #t 
		      "catch-errors caught exception ~s for cell ~s~%" 
		      (error-string condition)
		      ?cell))
	(exception ?cell (format #f "~a" (error-string condition))))
      (lambda () ?form ...)))
) )

(define-method lookup-name ( (name <symbol>) )
"returns value or #f"
  (with-error-handler
   (lambda (c) #f) ;; #f on fail
   (lambda ()
     (eval-at-top-level name (interaction-environment)))
) )


;;;==============================================
;;; class <date>
;;;==============================================
(define-class <date> ()
  ( (year    :accessor year    :init-keyword :year    
	     :type 'integer :init-form 0)
    (month   :accessor month   :init-keyword :month   
	     :type 'integer :init-form 0)
    (day     :accessor day     :init-keyword :day     
	     :type 'integer :init-form 0)
    (hours   :accessor hours   :init-keyword :hours   
	     :type 'integer :init-form 0)
    (minutes :accessor minutes :init-keyword :minutes 
	     :type 'integer :init-form 0)
    (seconds :accessor seconds :init-keyword :seconds 
	     :type 'integer :init-form 0)
) )

(define-method clone ( (d <date>) )
  (make <date>
    :year    (year  d)
    :month   (month d)
    :day     (day   d)
    :hours   (hours d)
    :minutes (minutes d)
    :seconds (seconds d)
) )

(define-method + ( (d1 <date>) (d2 <date>) )
  ;;@@@ FIXME time only
  (let* ( (s1 (+ (hours->seconds   (hours d1))
		 (minutes->seconds (minutes d1))
		 (seconds d1)))
	  (s2 (+ (hours->seconds   (hours d2))
		 (minutes->seconds (minutes d2))
		 (seconds d2)))
	  (delta (+ s1 s2))
	  (hours (inexact->exact (quotient  delta (* 60 60))))
	  (rem   (inexact->exact (remainder delta (* 60 60))))
	  (minutes (inexact->exact (quotient  rem 60)))
	  (seconds (remainder rem 60))
	)
       (make <date>
	 :year  (max (year d1)  (year d2))
	 :month (max (month d1) (month d2))
	 :day   (+   (day d1)   (day d2))
	 :hours hours :minutes minutes :seconds seconds)
) )

(define-method - ( (d1 <date>) (d2 <date>) )
  ;;@@@ FIXME time only
  (let* ( (s1 (+ (hours->seconds (hours d1))
		 (minutes->seconds (minutes d1))
		 (seconds d1)))
	  (s2 (+ (hours->seconds (hours d2))
		 (minutes->seconds (minutes d2))
		 (seconds d2)))
	  (delta (- s1 s2))
	  (hours (inexact->exact (quotient delta (* 60 60))))
	  (rem   (inexact->exact (remainder delta (* 60 60))))
	  (minutes (inexact->exact (quotient  rem 60)))
	  (seconds (remainder rem 60))
	)
       (make <date>
	 :hours hours :minutes minutes :seconds seconds)
) )

(define-method write-object ( (d <date>) port )
;;  (format port "#[<date>D/M/Y h:m AP=~a]"
  (format port "~a" (date->string d)))

(define-method string->date ( (str <string>) )
  ;; sample format: "9/9/02 9:01 AM" D/M/Y h:m AP
  ;; @@ FIXME -- not forgiving at all
 (let* ( (str         (trim-whitespace str))
	 (day-end     (search-for "/" str 0))
	 (month-end   (and day-end 
			   (search-for "/" str (add1 day-end))))
	 (year-end    (and month-end
			   (search-for " " str (add1 month-end))))
	 (hours-end   (and year-end
			   (search-for ":" str (add1 year-end))))
	 (minutes-end (and hours-end
			   (search-for " " str (add1 hours-end))))
	 (am/pm       (and minutes-end
			   (string-ref str (+ minutes-end 1))))
	)
   (unless minutes-end
	   (error (format #f "string->date: could not parse: ~s" str)))
;    (when *debug-fit*
; 	 (format #t "d#~a, m#~a, y#~a, h#~a, m#~a, ~am~%" 
; 		 day-end month-end year-end 
; 		 hours-end minutes-end 
; 		 am/pm))
   (let ( (day     (string->number (substring str 0 day-end)))
	  (month   (string->number 
		    (substring str (add1 day-end) month-end)))
	  (year    (string->number 
		    (substring str (add1 month-end) year-end)))
	  (hours   (string->number 
		    (substring str (add1 year-end) hours-end)))
	  (minutes (string->number 
		    (substring str (add1 hours-end) minutes-end)))
	)
;      (when *debug-fit*
; 	 (format #t "d=~a, m=~a, y=~a, h=~a, m=~a, ~am~%" 
; 		 day month year 
; 		 hours minutes 
; 		 am/pm))
     (when (memq am/pm '(#\p #\P))
	   (set! hours (+ hours 12)))
     (make <date> :year year :month month :day day 
	          :hours hours :minutes minutes)
) ) )

(define-method date->string ( (d <date>) )
;; D/M/Y h:m AP
  (let* ( (month   (month d))
	  (day     (day d))
	  (year    (year d))
	  (display-year ;; NB: 2 digits
	   (format #f (if (< year 10) "0~a" "~a") year))
	  (minutes (minutes d))
	  (hours   (hours d))
	  (am/pm 
	   (if (> hours 12) #\P #\A))
	  (display-hours 
	   (if (> hours 12) (- hours 12) hours))
        )
  (format #f "~a/~a/~a ~a:~a~a ~aM" 
	  day month display-year display-hours 
	  (if (< minutes 10) "0" "") minutes am/pm)
) )

(define-method date->hms ( (d <date>) )
  ;; hh:mm:ss
  (let ( (h (number->string (hours   d)))
	 (m (number->string (minutes d)))
	 (s (number->string (inexact->exact (seconds d))))
	 (digitize 
	   (lambda (s)
	     (if (< (string-length s) 2)
		 (string-append "0" s)
		 s))
	 )
       )
    (format #f "~a:~a:~a"
	    (digitize h)
	    (digitize m)
	    (digitize s))
) )

;;; --- E O F --- ;;;
