;;;;   -*- Mode: lisp; Package: common-lisp-user; Syntax: Common-lisp -*-
;;
;; Copyright (C) 2000-2001 Memetrics Pty. Ltd.
;; This code may be used for any purpose but without warrenty or liability
;; of any kind.
;;
;;;;   Implements: A simple unit test harness package.
;;;;   Author: Alain Picard, Memetrics Pty.
;;;;   File: $Id: unit-testing.lisp,v 1.1 2002/10/15 05:11:32 kend Exp $

;;;;   Modified for generic unit testing
;;;;	29 September 2001 by Ken Dickey.

;;;;
;;;;   Commentary
;;;;   ----------
;;;;
;;;;   For simple tests, users are expected to use the DEFTEST macro,
;;;;   which creates an instance of a new test which will run its body.
;;;;   When all the DEFTESTs in a file have been seen, the tests can be
;;;;   run by evaluating (run-all-tests).  Tests may also be run individually,
;;;;   if the test object which result from DEFTEST is stored and given as argument 
;;;;   to the RUN-TEST method.
;;;;
;;;;   The user writes forms in the DEFTEST which end up evaluating to test-fails,
;;;;   test-passes (possibly via the handy test-assert macro.  e.g.
;;;;   (deftest :unit-name "Test name"
;;;;      (let ((foo (make-bar)))
;;;;         (big-long-calc foo)
;;;;         (test-assert (some-condition foo) "Foo failed---wrong phase of moon")
;;;;
;;;;         ... more code here ...
;;;;
;;;;         (test-passes)))
;;;;
;;;;         
;;;;   For more complex tests requiring fancy setting up and tearing down
;;;;   (as well as reclamation of resources in case a test fails), users are expected
;;;;   to create a subclass of the unit-test class using the DEFINE-TEST-CLASS macro.
;;;;   The syntax is meant to be reminiscent of CLOS, e.g
;;;;   
;;;;   (define-test-class my-test-class
;;;;     ((my-slot-1 :initarg :foo ...)
;;;;      (my-slot-2 (any valid CLOS slot options)
;;;;      ....))
;;;;   After this, the methods
;;;;   (defgeneric run-test :before
;;;;               ((test my-test-class) &key (output *debug-io*))   and
;;;;   (defgeneric run-test :after
;;;;               ((test my-test-class) &key (output *debug-io*))   may be
;;;;   specialized to perform the required actions, possibly accessing the
;;;;   my-slot-1's, etc.
;;;;
;;;;   The test form is protected by a handler case.  Care should be taken
;;;;   than any run-test specialization also be protected not to crash.
;;;;
;;;;   You can run tests for a specific test unit via
;;;;	(run-all-tests :unit :unit-name)
;;;;   Of course (run-all-tests) runs all the tests.
;;;;
;;;;   See file "testing-tests.lisp" for sample usage.
;;;;
;;;;   Oh, yes.  In the odd chance that some code breaks,
;;;;   (setq *unit-test-debug* t) will BREAK at point of
;;;;   failure for optional debugging. 

;;;;
;;;;   We are busy folks and can't commit to update, or even respond,
;;;;   to enhancements and bug reports but we _are_ interested and may
;;;;   help as we get time:
;;;;
;;;;	Ken Dickey:   kend@memetrics.com, kend0@earthlink.net
;;;;    Alain Picard: Alain.Picard@memetrics.com
;;;;	bugs@memetrics.com
;;;;

;;;;
;;;; In defsystem:
;;;;	(make-package :<this-package>)
;;;;

;; Set to whatever package makes sense for <this-package>
(in-package :common-lisp-user)  

;; Typical Conventions:
;;   (defconstant +global+   ..)
;;   (defvar      *global*   ..)

(defconstant +test-success-msg+     "~%Test <<~A>>~60TPASSES.~&")
(defconstant +test-fail-msg+        "~%Test <<~A>>~60TFAILS.~&Failed on form: ~A~%")
(defconstant +test-assert-fail-msg+ "~%Test <<~A>>~60TFAILS.~&Test Assert Failed on form: ~A~%")
(defconstant +test-ensure-exception-raised-fail-msg+ "~%Test <<~A>>~60TFAILS.~
                                                      ~&Ensure Exception Raised Failed on form: ~A~%")
(defconstant +test-crash-msg-debug-off+ "~%Test <<~A>>~60TCRASHES.~
                                         ~&Test crashed on form: ~A~
                                         ~&~A
                                         ~&** To break on errors (setq *unit-test-debug* t).~%~%")

(defconstant +test-crash-msg-debug-on+ "~%Test <<~A>>~60TCRASHES.~
                                        ~&Test crashed on form: ~A~%")
  
(define-condition unit-test-failure ()
  ((reason :reader failure-reason
	   :initform 'unknown
	   :initarg :reason)
   (form   :reader form
	   :initarg :form)))

(define-condition test-assert-failure (unit-test-failure)
  ())

;;(define-condition plain-form-failure (unit-test-failure)
;;  ())

(define-condition ensure-exception-raised-failure (unit-test-failure)
  ())

(defvar *unit-tests* nil
  "This contains all unit tests.  Tests are added
simply by creating instances of test objects.")

(defvar *unit-test-debug* nil
  "(Setq *unit-test-debug* t) for debugging")

(defun test-fails (&optional (reason 'no-reason-given))
  "Used to signal that the test has failed.  REASON may be
a string or symbol which will be noted in the test report."
  (error 'unit-test-failure :reason reason))

(defun test-passes ()
  "Used to signal that the test has passed."
  t)

(defclass unit-test ()
  ((unit :reader unit-test-unit
	 :initarg :unit
	 :documentation "Name of unit being tested.
Used to group related tests.")
   (name :reader unit-test-name
	 :initarg :name
	 :initform "Unknown test")
   (forms :initarg :forms
	  :reader unit-test-forms)
   (failed-form :accessor unit-test-failed-form
		:initform "Unknown form")
   )
  (:documentation "The base class for all unit tests.  When tests are defined,
they will be stored in a global list of tests which can be run at any time with
RUN-ALL-TESTS."))

(defmethod print-object ((self unit-test) stream)
  (print-unreadable-object (self stream :type t :identity t)
     (format stream "~s ~(~s~)" (unit-test-name self) (unit-test-unit self))))

(defmethod initialize-instance :after ((self unit-test)
				       &rest rest
				       &key &allow-other-keys)
  ;; If (deftest ...) is called repeatedly, replace
  ;; the test by the same name, if possible.
  (let ((found? (position (unit-test-name self) *unit-tests* :test #'string= :key #'unit-test-name)))
    (if found?
	;; We need to replace the test
	 (setf (nth found? *unit-tests*) self)
	 ;; new test.
	 (push self *unit-tests*))))

(defun run-all-tests (&key (verbose t) (output *standard-output*) (unit :all))
  ;; Take care to run the tests in the order
  ;; in which they were added.
  (declare (ignore verbose))
  (flet ((eq-or-all (element element-unit-set)
	   (or (eq element element-unit-set) (eq element-unit-set :all))))
    (let ((unit-set (if (listp unit)
			unit
			(list unit)))
	  (results '()))
      
      ;; Run each test and collect result
      (dolist (test (reverse *unit-tests*))
	(when (member (unit-test-unit test) unit-set :test #'eq-or-all)
	  (push (run-test test :output output) results)))
      
      ;; Reporting
      (let ((ntotal (length results))
	    (npass  (count :passes  results))
	    (nfail  (count :fails   results))
	    (ncrash (count :crashes results)))
	(format output   "~&~%====================================")
	(format output   "~&~D tests total." ntotal)
	(format output   "~&   ~D passed."   npass)
	(when (not (zerop nfail))
	  (format output "~&   ~D failed."   nfail))
	(when (not (zerop ncrash))
	  (format output "~&   ~D crashed."  ncrash)))
      (finish-output output))))

(defun delete-all-tests ()
  "Throw away all tests from test suite."
  (setq *unit-tests* nil))

(defun delete-tests-for (unit)
  "Blow away all tests for UNIT from test set."
  (setq *unit-tests* (delete unit *unit-tests* :test #'eq :key #'unit-test-unit)))

(defmethod run-test ((self unit-test) &key (output *debug-io*))
  (handler-case
      (bind-when (result (evaluate-test self output)) ; Success
		 (format output +test-success-msg+ (unit-test-name self))
		 (finish-output output)
		 :passes)
    (unit-test-failure (cause)
      (format output (etypecase cause
		       (test-assert-failure +test-assert-fail-msg+)
		       (ensure-exception-raised-failure +test-ensure-exception-raised-fail-msg+)
;		       (plain-form-failure +test-fail-msg+)
		       (unit-test-failure +test-fail-msg+)
		       )
	      (unit-test-name self) (form cause))
      (finish-output output)
      :fails)
    
    (serious-condition (cause)		; Form is invalid, somehow
      (when (not *unit-test-debug*)
	(format output +test-crash-msg-debug-off+
		(unit-test-name self)
		(unit-test-failed-form self)
		cause)
	(finish-output output))
      :crashes)))

(defun evaluate-test (unit-test output)
  (check-type unit-test unit-test)
  
  (mapc #'(lambda (form)
	    (handler-bind
		((unit-test-failure
		  #'(lambda (c) ; Handle this condition in run-test
		      (error c)))
		 (serious-condition
		  #'(lambda (c)	; Crash. Unknown
		      ;;@@ TEST
		      (format *error-output* "~&~A~&" c)
		      ;;@@ TEST
		      (if *unit-test-debug*
			  (restart-case
			      (progn (format output +test-crash-msg-debug-on+
					     (unit-test-name unit-test)
					     form)
				     (finish-output output)
				     (invoke-debugger c))
			       (ignore&continue ()
				 :report (lambda (stream)
					   (format stream "Ignore & Continue testing.")
					   (finish-output stream))
				 (error c)))
			  (progn
			    (setf (unit-test-failed-form unit-test) form)
			    (error c))))))
	      (funcall (coerce `(lambda () ,form) 'function))))
	(unit-test-forms unit-test)))

;; Creating user-specific test which need special setting up/teardown actions
;; 
(defmacro define-test-class (class-name &optional slots)
  "Define a new test class.  SLOTS are specified as normal CLOS slots."
  `(defclass ,class-name (unit-test)
     ,slots))

(defmacro test-assert (form &optional (reason nil))
  (let ((result-sym (gensym "RESULT")))
    `(bind-if (,result-sym ,form)	
	     ,result-sym
             (error 'test-assert-failure :form ',form
	      :reason ,(or reason "No reason specified")))))

(defmacro ensure-exception-raised (&key form expected-condition assertion addendum)
  "Use only within the scope of deftest."
  (let ((returned-value-sym (gensym "VALUES-LIST")))
    `(handler-case (let ((,returned-value-sym ,form))
		     (error 'ensure-exception-raised-failure :form ',form
			    :reason (format nil "~&No exception raised while processing form :~%~
                                                 ~W~% -> ~W~%  expected a condition of type ~a."
					    ',form ,returned-value-sym
					    ',expected-condition)))
       (,expected-condition (condition-raised) ; note 'condition-raised' variable here!
	  (declare (ignorable condition-raised))
	  (test-assert ,assertion ,addendum)))))

(defmacro deftest (unit name &rest body)
  "Create an instance of a unit test called NAME for UNIT."
  `(make-instance 'unit-test
     :unit ,unit
     :name ,name
     :forms ',body
    ))

#|
(defparameter *test-initialize-subsystems-done?* nil
  "Set to nil to force a re-initialization")

(defun test-initialize-subsystems (&optional (force? nil))
  "Set up the environment so as to be able to run some CORBA external tests."
  (when (or force? (not *test-initialize-subsystems-done?*))
    (let ((args-info (grovel-over-command-line-arguments)))
      (initialize-logging-subsystem (funcall args-info "--log"))
      (initialize-corba-subsystem)
      (initialize-configuration     (funcall args-info "--conf"))
      (initialize-db-subsystem)
      (initialize-object-database)))
  (setq *test-initialize-subsystems-done?* t))
|#

;;; UNIT-TESTING.LISP ends here
